"""Tests for cwltool.load_tool."""

import logging
import urllib.parse
from pathlib import Path

import pytest
from schema_salad.exceptions import ValidationException

from cwltool.context import LoadingContext, RuntimeContext
from cwltool.errors import WorkflowException
from cwltool.load_tool import load_tool
from cwltool.loghandler import _logger, configure_logging
from cwltool.process import use_custom_schema, use_standard_schema
from cwltool.update import INTERNAL_VERSION
from cwltool.utils import CWLObjectType

from .util import get_data

configure_logging(_logger.handlers[-1], False, False, True, True, True)
_logger.setLevel(logging.DEBUG)


def test_check_version() -> None:
    """
    It is permitted to load without updating, but not execute.

    Attempting to execute without updating to the internal version should raise an error.
    """
    joborder: CWLObjectType = {"inp": "abc"}
    loadingContext = LoadingContext({"do_update": True})
    tool = load_tool(get_data("tests/echo.cwl"), loadingContext)
    for _ in tool.job(joborder, None, RuntimeContext()):
        pass

    loadingContext = LoadingContext({"do_update": False})
    tool = load_tool(get_data("tests/echo.cwl"), loadingContext)
    with pytest.raises(WorkflowException):
        for _ in tool.job(joborder, None, RuntimeContext()):
            pass


def test_use_metadata() -> None:
    """Use the version from loadingContext.metadata if cwlVersion isn't present in the document."""
    loadingContext = LoadingContext({"do_update": False})
    tool = load_tool(get_data("tests/echo.cwl"), loadingContext)

    loadingContext = LoadingContext()
    loadingContext.metadata = tool.metadata
    tooldata = tool.tool.copy()
    del tooldata["cwlVersion"]
    load_tool(tooldata, loadingContext)


def test_checklink_outputSource() -> None:
    """Is outputSource resolved correctly independent of value of do_validate."""
    outsrc = Path(get_data("tests/wf/1st-workflow.cwl")).as_uri() + "#argument/classfile"

    loadingContext = LoadingContext({"do_validate": True})
    tool = load_tool(get_data("tests/wf/1st-workflow.cwl"), loadingContext)
    assert tool.tool["outputs"][0]["outputSource"] == outsrc

    loadingContext = LoadingContext({"do_validate": False})
    tool = load_tool(get_data("tests/wf/1st-workflow.cwl"), loadingContext)
    assert tool.tool["outputs"][0]["outputSource"] == outsrc


def test_load_graph_fragment() -> None:
    """Reloading from a dictionary without a cwlVersion."""
    loadingContext = LoadingContext()
    uri = Path(get_data("tests/wf/scatter-wf4.cwl")).as_uri() + "#main"
    tool = load_tool(uri, loadingContext)

    loader = tool.doc_loader
    assert loader
    rs, metadata = loader.resolve_ref(uri)
    # Reload from a dict (in 'rs'), not a URI.  The dict is a fragment
    # of original document and doesn't have cwlVersion set, so test
    # that it correctly looks up the root document to get the
    # cwlVersion.
    assert isinstance(rs, str) or isinstance(rs, dict)
    tool = load_tool(rs, loadingContext)
    assert tool.metadata["cwlVersion"] == INTERNAL_VERSION


def test_load_graph_fragment_from_packed() -> None:
    """Loading a fragment from packed with update."""
    loadingContext = LoadingContext()
    uri = Path(get_data("tests/wf/packed-with-loadlisting.cwl")).as_uri() + "#main"
    try:
        with open(get_data("extensions.yml")) as res:
            use_custom_schema("v1.0", "http://commonwl.org/cwltool", res.read())

        # The updater transforms LoadListingRequirement from an
        # extension (in v1.0) to a core feature (in v1.1) but there
        # was a bug when loading a packed workflow and loading a
        # specific fragment it would get the un-updated document.
        # This recreates that case and asserts that we are using the
        # updated document like we should.

        tool = load_tool(uri, loadingContext)

        assert tool.tool["requirements"] == [
            {"class": "LoadListingRequirement", "loadListing": "no_listing"}
        ]

        # This tests an additional, related bug, in which the updater
        # updates only one fragment of the document, but would update
        # the metadata dict in-place.  This had the effect of marking
        # the original document as updated.  On a subsequent load, it
        # would get the original un-updated document, which mistakenly
        # had the version modified in place, and validate it with the
        # newer schema instead of the original one.
        #
        # The specific case where this failed was
        # cwltool:LoadListingRequirement which is only recognized for
        # v1.0 documents, newer documents are supposed to be
        # auto-updated to the spec LoadListingRequirement, so it is
        # dropped from the schema.  However if we try to validate a
        # v1.0 document with a v1.2 schema, it will throw an error the
        # cwltool:LoadListingRequirement extension.
        #
        # This was solved by making a shallow copy of the metadata
        # dict to ensure that the updater did not modify the original
        # document.
        uri2 = Path(get_data("tests/wf/packed-with-loadlisting.cwl")).as_uri() + "#16169-step.cwl"
        load_tool(uri2, loadingContext)

    finally:
        use_standard_schema("v1.0")


def test_import_tracked() -> None:
    """Test that $import and $include are tracked in the index."""

    loadingContext = LoadingContext({"fast_parser": True})
    tool = load_tool(get_data("tests/wf/811-12.cwl"), loadingContext)
    path = f"import:file://{get_data('tests/wf/schemadef-type.yml')}"
    path2 = f"import:file://{urllib.parse.quote(get_data('tests/wf/schemadef-type.yml'))}"

    assert tool.doc_loader is not None
    assert path in tool.doc_loader.idx or path2 in tool.doc_loader.idx

    loadingContext = LoadingContext({"fast_parser": False})
    tool = load_tool(get_data("tests/wf/811.cwl"), loadingContext)

    assert tool.doc_loader is not None
    assert path in tool.doc_loader.idx or path2 in tool.doc_loader.idx


def test_load_badhints() -> None:
    """Check for expected error while update a bads hints list."""
    loadingContext = LoadingContext()
    uri = Path(get_data("tests/wf/hello-workflow-badhints.cwl")).as_uri()
    with pytest.raises(
        ValidationException,
        match=r".*tests\/wf\/hello-workflow-badhints\.cwl\:18:4:\s*'hints'\s*entry\s*missing\s*required\s*key\s*'class'\.",
    ):
        load_tool(uri, loadingContext)


def test_load_badhints_nodict() -> None:
    """Check for expected error while update a hints list with a numerical entry."""
    loadingContext = LoadingContext()
    uri = Path(get_data("tests/wf/hello-workflow-badhints2.cwl")).as_uri()
    with pytest.raises(
        ValidationException,
        match=r".*tests\/wf\/hello-workflow-badhints2\.cwl:41:5:\s*'hints'\s*entries\s*must\s*be\s*dictionaries:\s*<class\s*'int'>\s*42\.",
    ):
        load_tool(uri, loadingContext)



import copy
import datetime
import logging
import urllib
import uuid
from io import BytesIO
from pathlib import PurePath, PurePosixPath
from socket import getfqdn
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from prov.identifier import Identifier, QualifiedName
from prov.model import PROV, PROV_LABEL, PROV_TYPE, PROV_VALUE, ProvDocument, ProvEntity
from schema_salad.sourceline import SourceLine

from ..errors import WorkflowException
from ..job import CommandLineJob, JobBase
from ..loghandler import _logger
from ..process import Process, shortname
from ..stdfsaccess import StdFsAccess
from ..utils import CWLObjectType, JobsType, get_listing, posix_path, versionstring
from ..workflow_job import WorkflowJob
from .provenance_constants import (
    ACCOUNT_UUID,
    CWLPROV,
    ENCODING,
    FOAF,
    METADATA,
    ORE,
    PROVENANCE,
    RO,
    SCHEMA,
    SHA1,
    SHA256,
    TEXT_PLAIN,
    UUID,
    WF4EVER,
    WFDESC,
    WFPROV,
)
from .writablebagfile import create_job, write_bag_file  # change this later

if TYPE_CHECKING:
    from .ro import ResearchObject


def copy_job_order(job: Union[Process, JobsType], job_order_object: CWLObjectType) -> CWLObjectType:
    """Create copy of job object for provenance."""
    if not isinstance(job, WorkflowJob):
        # direct command line tool execution
        return job_order_object
    customised_job: CWLObjectType = {}
    # new job object for RO
    debug = _logger.isEnabledFor(logging.DEBUG)
    for each, i in enumerate(job.tool["inputs"]):
        with SourceLine(job.tool["inputs"], each, WorkflowException, debug):
            iid = shortname(i["id"])
            if iid in job_order_object:
                customised_job[iid] = copy.deepcopy(job_order_object[iid])
                # add the input element in dictionary for provenance
            elif "default" in i:
                customised_job[iid] = copy.deepcopy(i["default"])
                # add the default elements in the dictionary for provenance
            else:
                pass
    return customised_job


class ProvenanceProfile:
    """
    Provenance profile.

    Populated as the workflow runs.
    """

    def __init__(
        self,
        research_object: "ResearchObject",
        full_name: str,
        host_provenance: bool,
        user_provenance: bool,
        orcid: str,
        fsaccess: StdFsAccess,
        run_uuid: Optional[uuid.UUID] = None,
    ) -> None:
        """Initialize the provenance profile."""
        self.fsaccess = fsaccess
        self.orcid = orcid
        self.research_object = research_object
        self.folder = self.research_object.folder
        self.document = ProvDocument()
        self.host_provenance = host_provenance
        self.user_provenance = user_provenance
        self.engine_uuid = research_object.engine_uuid
        self.add_to_manifest = self.research_object.add_to_manifest
        if self.orcid:
            _logger.debug("[provenance] Creator ORCID: %s", self.orcid)
        self.full_name = full_name
        if self.full_name:
            _logger.debug("[provenance] Creator Full name: %s", self.full_name)
        self.workflow_run_uuid = run_uuid or uuid.uuid4()
        self.workflow_run_uri = self.workflow_run_uuid.urn
        self.generate_prov_doc()

    def __str__(self) -> str:
        """Represent this Provenvance profile as a string."""
        return f"ProvenanceProfile <{self.workflow_run_uri}> in <{self.research_object}>"

    def generate_prov_doc(self) -> Tuple[str, ProvDocument]:
        """Add basic namespaces."""

        def host_provenance(document: ProvDocument) -> None:
            """Record host provenance."""
            document.add_namespace(CWLPROV)
            document.add_namespace(UUID)
            document.add_namespace(FOAF)

            hostname = getfqdn()
            # won't have a foaf:accountServiceHomepage for unix hosts, but
            # we can at least provide hostname
            document.agent(
                ACCOUNT_UUID,
                {
                    PROV_TYPE: FOAF["OnlineAccount"],
                    "prov:location": hostname,
                    CWLPROV["hostname"]: hostname,
                },
            )

        self.cwltool_version = f"cwltool {versionstring().split()[-1]}"
        self.document.add_namespace("wfprov", "http://purl.org/wf4ever/wfprov#")
        # document.add_namespace('prov', 'http://www.w3.org/ns/prov#')
        self.document.add_namespace("wfdesc", "http://purl.org/wf4ever/wfdesc#")
        # TODO: Make this ontology. For now only has cwlprov:image
        self.document.add_namespace("cwlprov", "https://w3id.org/cwl/prov#")
        self.document.add_namespace("foaf", "http://xmlns.com/foaf/0.1/")
        self.document.add_namespace("schema", "http://schema.org/")
        self.document.add_namespace("orcid", "https://orcid.org/")
        self.document.add_namespace("id", "urn:uuid:")
        # NOTE: Internet draft expired 2004-03-04 (!)
        #  https://tools.ietf.org/html/draft-thiemann-hash-urn-01
        # TODO: Change to nih:sha-256; hashes
        #  https://tools.ietf.org/html/rfc6920#section-7
        self.document.add_namespace("data", "urn:hash::sha1:")
        # Also needed for docker images
        self.document.add_namespace(SHA256, "nih:sha-256;")

        # info only, won't really be used by prov as sub-resources use /
        self.document.add_namespace("researchobject", self.research_object.base_uri)
        # annotations
        self.metadata_ns = self.document.add_namespace(
            "metadata", self.research_object.base_uri + METADATA + "/"
        )
        # Pre-register provenance directory so we can refer to its files
        self.provenance_ns = self.document.add_namespace(
            "provenance", self.research_object.base_uri + posix_path(PROVENANCE) + "/"
        )
        ro_identifier_workflow = self.research_object.base_uri + "workflow/packed.cwl#"
        self.wf_ns = self.document.add_namespace("wf", ro_identifier_workflow)
        ro_identifier_input = self.research_object.base_uri + "workflow/primary-job.json#"
        self.document.add_namespace("input", ro_identifier_input)

        # More info about the account (e.g. username, fullname)
        # may or may not have been previously logged by user_provenance()
        # .. but we always know cwltool was launched (directly or indirectly)
        # by a user account, as cwltool is a command line tool
        account = self.document.agent(ACCOUNT_UUID)
        if self.orcid or self.full_name:
            person: Dict[Union[str, Identifier], Any] = {
                PROV_TYPE: PROV["Person"],
                "prov:type": SCHEMA["Person"],
            }
            if self.full_name:
                person["prov:label"] = self.full_name
                person["foaf:name"] = self.full_name
                person["schema:name"] = self.full_name
            else:
                # TODO: Look up name from ORCID API?
                pass
            agent = self.document.agent(self.orcid or uuid.uuid4().urn, person)
            self.document.actedOnBehalfOf(account, agent)
        else:
            if self.host_provenance:
                host_provenance(self.document)
            if self.user_provenance:
                self.research_object.user_provenance(self.document)
        # The execution of cwltool
        wfengine = self.document.agent(
            self.engine_uuid,
            {
                PROV_TYPE: PROV["SoftwareAgent"],
                "prov:type": WFPROV["WorkflowEngine"],
                "prov:label": self.cwltool_version,
            },
        )
        # FIXME: This datetime will be a bit too delayed, we should
        # capture when cwltool.py earliest started?
        self.document.wasStartedBy(wfengine, None, account, datetime.datetime.now())
        # define workflow run level activity
        self.document.activity(
            self.workflow_run_uri,
            datetime.datetime.now(),
            None,
            {
                PROV_TYPE: WFPROV["WorkflowRun"],
                "prov:label": "Run of workflow/packed.cwl#main",
            },
        )
        # association between SoftwareAgent and WorkflowRun
        main_workflow = "wf:main"
        self.document.wasAssociatedWith(self.workflow_run_uri, self.engine_uuid, main_workflow)
        self.document.wasStartedBy(
            self.workflow_run_uri, None, self.engine_uuid, datetime.datetime.now()
        )
        return (self.workflow_run_uri, self.document)

    def evaluate(
        self,
        process: Process,
        job: JobsType,
        job_order_object: CWLObjectType,
        research_obj: "ResearchObject",
    ) -> None:
        """Evaluate the nature of job."""
        if not hasattr(process, "steps"):
            # record provenance of independent commandline tool executions
            self.prospective_prov(job)
            customised_job = copy_job_order(job, job_order_object)
            self.used_artefacts(customised_job, self.workflow_run_uri)
            create_job(research_obj, customised_job)
        elif hasattr(job, "workflow"):
            # record provenance of workflow executions
            self.prospective_prov(job)
            customised_job = copy_job_order(job, job_order_object)
            self.used_artefacts(customised_job, self.workflow_run_uri)

    def record_process_start(
        self, process: Process, job: JobsType, process_run_id: Optional[str] = None
    ) -> Optional[str]:
        if not hasattr(process, "steps"):
            process_run_id = self.workflow_run_uri
        elif not hasattr(job, "workflow"):
            # commandline tool execution as part of workflow
            name = ""
            if isinstance(job, (CommandLineJob, JobBase, WorkflowJob)):
                name = job.name
            process_name = urllib.parse.quote(name, safe=":/,#")
            process_run_id = self.start_process(process_name, datetime.datetime.now())
        return process_run_id

    def start_process(
        self,
        process_name: str,
        when: datetime.datetime,
        process_run_id: Optional[str] = None,
    ) -> str:
        """Record the start of each Process."""
        if process_run_id is None:
            process_run_id = uuid.uuid4().urn
        prov_label = "Run of workflow/packed.cwl#main/" + process_name
        self.document.activity(
            process_run_id,
            None,
            None,
            {PROV_TYPE: WFPROV["ProcessRun"], PROV_LABEL: prov_label},
        )
        self.document.wasAssociatedWith(
            process_run_id, self.engine_uuid, str("wf:main/" + process_name)
        )
        self.document.wasStartedBy(process_run_id, None, self.workflow_run_uri, when, None, None)
        return process_run_id

    def record_process_end(
        self,
        process_name: str,
        process_run_id: str,
        outputs: Union[CWLObjectType, MutableSequence[CWLObjectType], None],
        when: datetime.datetime,
    ) -> None:
        self.generate_output_prov(outputs, process_run_id, process_name)
        self.document.wasEndedBy(process_run_id, None, self.workflow_run_uri, when)

    def declare_file(self, value: CWLObjectType) -> Tuple[ProvEntity, ProvEntity, str]:
        if value["class"] != "File":
            raise ValueError("Must have class:File: %s" % value)
        # Need to determine file hash aka RO filename
        entity: Optional[ProvEntity] = None
        checksum = None
        if "checksum" in value:
            csum = cast(str, value["checksum"])
            (method, checksum) = csum.split("$", 1)
            if method == SHA1 and self.research_object.has_data_file(checksum):
                entity = self.document.entity("data:" + checksum)

        if not entity and "location" in value:
            location = str(value["location"])
            # If we made it here, we'll have to add it to the RO
            with self.fsaccess.open(location, "rb") as fhandle:
                relative_path = self.research_object.add_data_file(fhandle)
                # FIXME: This naively relies on add_data_file setting hash as filename
                checksum = PurePath(relative_path).name
                entity = self.document.entity("data:" + checksum, {PROV_TYPE: WFPROV["Artifact"]})
                if "checksum" not in value:
                    value["checksum"] = f"{SHA1}${checksum}"

        if not entity and "contents" in value:
            # Anonymous file, add content as string
            entity, checksum = self.declare_string(cast(str, value["contents"]))

        # By here one of them should have worked!
        if not entity or not checksum:
            raise ValueError("class:File but missing checksum/location/content: %r" % value)

        # Track filename and extension, this is generally useful only for
        # secondaryFiles. Note that multiple uses of a file might thus record
        # different names for the same entity, so we'll
        # make/track a specialized entity by UUID
        file_id = cast(str, value.setdefault("@id", uuid.uuid4().urn))
        # A specialized entity that has just these names
        file_entity = self.document.entity(
            file_id,
            [(PROV_TYPE, WFPROV["Artifact"]), (PROV_TYPE, WF4EVER["File"])],
        )

        if "basename" in value:
            file_entity.add_attributes({CWLPROV["basename"]: cast(str, value["basename"])})
        if "nameroot" in value:
            file_entity.add_attributes({CWLPROV["nameroot"]: cast(str, value["nameroot"])})
        if "nameext" in value:
            file_entity.add_attributes({CWLPROV["nameext"]: cast(str, value["nameext"])})
        self.document.specializationOf(file_entity, entity)

        # Check for secondaries
        for sec in cast(MutableSequence[CWLObjectType], value.get("secondaryFiles", [])):
            # TODO: Record these in a specializationOf entity with UUID?
            if sec["class"] == "File":
                (sec_entity, _, _) = self.declare_file(sec)
            elif sec["class"] == "Directory":
                sec_entity = self.declare_directory(sec)
            else:
                raise ValueError(f"Got unexpected secondaryFiles value: {sec}")
            # We don't know how/when/where the secondary file was generated,
            # but CWL convention is a kind of summary/index derived
            # from the original file. As its generally in a different format
            # then prov:Quotation is not appropriate.
            self.document.derivation(
                sec_entity,
                file_entity,
                other_attributes={PROV["type"]: CWLPROV["SecondaryFile"]},
            )

        return file_entity, entity, checksum

    def declare_directory(self, value: CWLObjectType) -> ProvEntity:
        """Register any nested files/directories."""
        # FIXME: Calculate a hash-like identifier for directory
        # so we get same value if it's the same filenames/hashes
        # in a different location.
        # For now, mint a new UUID to identify this directory, but
        # attempt to keep it inside the value dictionary
        dir_id = cast(str, value.setdefault("@id", uuid.uuid4().urn))

        # New annotation file to keep the ORE Folder listing
        ore_doc_fn = dir_id.replace("urn:uuid:", "directory-") + ".ttl"
        dir_bundle = self.document.bundle(self.metadata_ns[ore_doc_fn])

        coll = self.document.entity(
            dir_id,
            [
                (PROV_TYPE, WFPROV["Artifact"]),
                (PROV_TYPE, PROV["Collection"]),
                (PROV_TYPE, PROV["Dictionary"]),
                (PROV_TYPE, RO["Folder"]),
            ],
        )

        if "basename" in value:
            coll.add_attributes({CWLPROV["basename"]: cast(str, value["basename"])})

        # ORE description of ro:Folder, saved separately
        coll_b = dir_bundle.entity(
            dir_id,
            [(PROV_TYPE, RO["Folder"]), (PROV_TYPE, ORE["Aggregation"])],
        )
        self.document.mentionOf(dir_id + "#ore", dir_id, dir_bundle.identifier)

        # dir_manifest = dir_bundle.entity(
        #     dir_bundle.identifier, {PROV["type"]: ORE["ResourceMap"],
        #                             ORE["describes"]: coll_b.identifier})

        coll_attribs: List[Tuple[Union[str, Identifier], Any]] = [
            (ORE["isDescribedBy"], dir_bundle.identifier)
        ]
        coll_b_attribs: List[Tuple[Union[str, Identifier], Any]] = []

        # FIXME: .listing might not be populated yet - hopefully
        # a later call to this method will sort that
        is_empty = True

        if "listing" not in value:
            get_listing(self.fsaccess, value)
        for entry in cast(MutableSequence[CWLObjectType], value.get("listing", [])):
            is_empty = False
            # Declare child-artifacts
            entity = self.declare_artefact(entry)
            self.document.membership(coll, entity)
            # Membership relation aka our ORE Proxy
            m_id = uuid.uuid4().urn
            m_entity = self.document.entity(m_id)
            m_b = dir_bundle.entity(m_id)

            # PROV-O style Dictionary
            # https://www.w3.org/TR/prov-dictionary/#dictionary-ontological-definition
            # ..as prov.py do not currently allow PROV-N extensions
            # like hadDictionaryMember(..)
            m_entity.add_asserted_type(PROV["KeyEntityPair"])

            m_entity.add_attributes(
                {
                    PROV["pairKey"]: cast(str, entry["basename"]),
                    PROV["pairEntity"]: entity,
                }
            )

            # As well as a being a
            # http://wf4ever.github.io/ro/2016-01-28/ro/#FolderEntry
            m_b.add_asserted_type(RO["FolderEntry"])
            m_b.add_asserted_type(ORE["Proxy"])
            m_b.add_attributes(
                {
                    RO["entryName"]: cast(str, entry["basename"]),
                    ORE["proxyIn"]: coll,
                    ORE["proxyFor"]: entity,
                }
            )
            coll_attribs.append((PROV["hadDictionaryMember"], m_entity))
            coll_b_attribs.append((ORE["aggregates"], m_b))

        coll.add_attributes(coll_attribs)
        coll_b.add_attributes(coll_b_attribs)

        # Also Save ORE Folder as annotation metadata
        ore_doc = ProvDocument()
        ore_doc.add_namespace(ORE)
        ore_doc.add_namespace(RO)
        ore_doc.add_namespace(UUID)
        ore_doc.add_bundle(dir_bundle)
        ore_doc = ore_doc.flattened()
        ore_doc_path = str(PurePosixPath(METADATA, ore_doc_fn))
        with write_bag_file(self.research_object, ore_doc_path) as provenance_file:
            ore_doc.serialize(provenance_file, format="rdf", rdf_format="turtle")
        self.research_object.add_annotation(dir_id, [ore_doc_fn], ORE["isDescribedBy"].uri)

        if is_empty:
            # Empty directory
            coll.add_asserted_type(PROV["EmptyCollection"])
            coll.add_asserted_type(PROV["EmptyDictionary"])
        self.research_object.add_uri(coll.identifier.uri)
        return coll

    def declare_string(self, value: str) -> Tuple[ProvEntity, str]:
        """Save as string in UTF-8."""
        byte_s = BytesIO(str(value).encode(ENCODING))
        data_file = self.research_object.add_data_file(byte_s, content_type=TEXT_PLAIN)
        checksum = PurePosixPath(data_file).name
        # FIXME: Don't naively assume add_data_file uses hash in filename!
        data_id = f"data:{PurePosixPath(data_file).stem}"
        entity = self.document.entity(
            data_id, {PROV_TYPE: WFPROV["Artifact"], PROV_VALUE: str(value)}
        )
        return entity, checksum

    def declare_artefact(self, value: Any) -> ProvEntity:
        """Create data artefact entities for all file objects."""
        if value is None:
            # FIXME: If this can happen in CWL, we'll
            # need a better way to represent this in PROV
            return self.document.entity(CWLPROV["None"], {PROV_LABEL: "None"})

        if isinstance(value, (bool, int, float)):
            # Typically used in job documents for flags

            # FIXME: Make consistent hash URIs for these
            # that somehow include the type
            # (so "1" != 1 != "1.0" != true)
            entity = self.document.entity(uuid.uuid4().urn, {PROV_VALUE: value})
            self.research_object.add_uri(entity.identifier.uri)
            return entity

        if isinstance(value, str):
            (entity, _) = self.declare_string(value)
            return entity

        if isinstance(value, bytes):
            # If we got here then we must be in Python 3
            byte_s = BytesIO(value)
            data_file = self.research_object.add_data_file(byte_s)
            # FIXME: Don't naively assume add_data_file uses hash in filename!
            data_id = f"data:{PurePosixPath(data_file).stem}"
            return self.document.entity(
                data_id,
                {PROV_TYPE: WFPROV["Artifact"], PROV_VALUE: str(value)},
            )

        if isinstance(value, MutableMapping):
            if "@id" in value:
                # Already processed this value, but it might not be in this PROV
                entities = self.document.get_record(value["@id"])
                if entities:
                    return cast(List[ProvEntity], entities)[0]
                # else, unknown in PROV, re-add below as if it's fresh

            # Base case - we found a File we need to update
            if value.get("class") == "File":
                (entity, _, _) = self.declare_file(value)
                value["@id"] = entity.identifier.uri
                return entity

            if value.get("class") == "Directory":
                entity = self.declare_directory(value)
                value["@id"] = entity.identifier.uri
                return entity
            coll_id = value.setdefault("@id", uuid.uuid4().urn)
            # some other kind of dictionary?
            # TODO: also Save as JSON
            coll = self.document.entity(
                coll_id,
                [
                    (PROV_TYPE, WFPROV["Artifact"]),
                    (PROV_TYPE, PROV["Collection"]),
                    (PROV_TYPE, PROV["Dictionary"]),
                ],
            )

            if value.get("class"):
                _logger.warning("Unknown data class %s.", value["class"])
                # FIXME: The class might be "http://example.com/somethingelse"
                coll.add_asserted_type(CWLPROV[value["class"]])

            # Let's iterate and recurse
            coll_attribs: List[Tuple[Union[str, Identifier], Any]] = []
            for key, val in value.items():
                v_ent = self.declare_artefact(val)
                self.document.membership(coll, v_ent)
                m_entity = self.document.entity(uuid.uuid4().urn)
                # Note: only support PROV-O style dictionary
                # https://www.w3.org/TR/prov-dictionary/#dictionary-ontological-definition
                # as prov.py do not easily allow PROV-N extensions
                m_entity.add_asserted_type(PROV["KeyEntityPair"])
                m_entity.add_attributes({PROV["pairKey"]: str(key), PROV["pairEntity"]: v_ent})
                coll_attribs.append((PROV["hadDictionaryMember"], m_entity))
            coll.add_attributes(coll_attribs)
            self.research_object.add_uri(coll.identifier.uri)
            return coll

        # some other kind of Collection?
        # TODO: also save as JSON
        try:
            members = []
            for each_input_obj in iter(value):
                # Recurse and register any nested objects
                e = self.declare_artefact(each_input_obj)
                members.append(e)

            # If we reached this, then we were allowed to iterate
            coll = self.document.entity(
                uuid.uuid4().urn,
                [
                    (PROV_TYPE, WFPROV["Artifact"]),
                    (PROV_TYPE, PROV["Collection"]),
                ],
            )
            if not members:
                coll.add_asserted_type(PROV["EmptyCollection"])
            else:
                for member in members:
                    # FIXME: This won't preserve order, for that
                    # we would need to use PROV.Dictionary
                    # with numeric keys
                    self.document.membership(coll, member)
            self.research_object.add_uri(coll.identifier.uri)
            # FIXME: list value does not support adding "@id"
            return coll
        except TypeError:
            _logger.warning("Unrecognized type %s of %r", type(value), value)
            # Let's just fall back to Python repr()
            entity = self.document.entity(uuid.uuid4().urn, {PROV_LABEL: repr(value)})
            self.research_object.add_uri(entity.identifier.uri)
            return entity

    def used_artefacts(
        self,
        job_order: Union[CWLObjectType, List[CWLObjectType]],
        process_run_id: str,
        name: Optional[str] = None,
    ) -> None:
        """Add used() for each data artefact."""
        if isinstance(job_order, list):
            for entry in job_order:
                self.used_artefacts(entry, process_run_id, name)
        else:
            # FIXME: Use workflow name in packed.cwl, "main" is wrong for nested workflows
            base = "main"
            if name is not None:
                base += "/" + name
            for key, value in job_order.items():
                prov_role = self.wf_ns[f"{base}/{key}"]
                try:
                    entity = self.declare_artefact(value)
                    self.document.used(
                        process_run_id,
                        entity,
                        datetime.datetime.now(),
                        None,
                        {"prov:role": prov_role},
                    )
                except OSError:
                    pass

    def generate_output_prov(
        self,
        final_output: Union[CWLObjectType, MutableSequence[CWLObjectType], None],
        process_run_id: Optional[str],
        name: Optional[str],
    ) -> None:
        """Call wasGeneratedBy() for each output,copy the files into the RO."""
        if isinstance(final_output, MutableSequence):
            for entry in final_output:
                self.generate_output_prov(entry, process_run_id, name)
        elif final_output is not None:
            # Timestamp should be created at the earliest
            timestamp = datetime.datetime.now()

            # For each output, find/register the corresponding
            # entity (UUID) and document it as generated in
            # a role corresponding to the output
            for output, value in final_output.items():
                entity = self.declare_artefact(value)
                if name is not None:
                    name = urllib.parse.quote(str(name), safe=":/,#")
                    # FIXME: Probably not "main" in nested workflows
                    role = self.wf_ns[f"main/{name}/{output}"]
                else:
                    role = self.wf_ns[f"main/{output}"]

                if not process_run_id:
                    process_run_id = self.workflow_run_uri

                self.document.wasGeneratedBy(
                    entity, process_run_id, timestamp, None, {"prov:role": role}
                )

    def prospective_prov(self, job: JobsType) -> None:
        """Create prospective prov recording as wfdesc prov:Plan."""
        if not isinstance(job, WorkflowJob):
            # direct command line tool execution
            self.document.entity(
                "wf:main",
                {
                    PROV_TYPE: WFDESC["Process"],
                    "prov:type": PROV["Plan"],
                    "prov:label": "Prospective provenance",
                },
            )
            return

        self.document.entity(
            "wf:main",
            {
                PROV_TYPE: WFDESC["Workflow"],
                "prov:type": PROV["Plan"],
                "prov:label": "Prospective provenance",
            },
        )

        for step in job.steps:
            stepnametemp = "wf:main/" + str(step.name)[5:]
            stepname = urllib.parse.quote(stepnametemp, safe=":/,#")
            provstep = self.document.entity(
                stepname,
                {PROV_TYPE: WFDESC["Process"], "prov:type": PROV["Plan"]},
            )
            self.document.entity(
                "wf:main",
                {
                    "wfdesc:hasSubProcess": provstep,
                    "prov:label": "Prospective provenance",
                },
            )
        # TODO: Declare roles/parameters as well

    def activity_has_provenance(self, activity: str, prov_ids: Sequence[Identifier]) -> None:
        """Add http://www.w3.org/TR/prov-aq/ relations to nested PROV files."""
        # NOTE: The below will only work if the corresponding metadata/provenance arcp URI
        # is a pre-registered namespace in the PROV Document
        attribs: List[Tuple[Union[str, Identifier], Any]] = [
            (PROV["has_provenance"], prov_id) for prov_id in prov_ids
        ]
        self.document.activity(activity, other_attributes=attribs)
        # Tip: we can't use https://www.w3.org/TR/prov-links/#term-mention
        # as prov:mentionOf() is only for entities, not activities
        uris = [i.uri for i in prov_ids]
        self.research_object.add_annotation(activity, uris, PROV["has_provenance"].uri)

    def finalize_prov_profile(self, name: Optional[str]) -> List[QualifiedName]:
        """Transfer the provenance related files to the RO."""
        # NOTE: Relative posix path
        if name is None:
            # main workflow, fixed filenames
            filename = "primary.cwlprov"
        else:
            # ASCII-friendly filename, avoiding % as we don't want %2520 in manifest.json
            wf_name = urllib.parse.quote(str(name), safe="").replace("%", "_")
            # Note that the above could cause overlaps for similarly named
            # workflows, but that's OK as we'll also include run uuid
            # which also covers thhe case of this step being run in
            # multiple places or iterations
            filename = f"{wf_name}.{self.workflow_run_uuid}.cwlprov"

        basename = str(PurePosixPath(PROVENANCE) / filename)

        # TODO: Also support other profiles than CWLProv, e.g. ProvOne

        # list of prov identifiers of provenance files
        prov_ids = []

        # https://www.w3.org/TR/prov-xml/
        with write_bag_file(self.research_object, basename + ".xml") as provenance_file:
            self.document.serialize(provenance_file, format="xml", indent=4)
            prov_ids.append(self.provenance_ns[filename + ".xml"])

        # https://www.w3.org/TR/prov-n/
        with write_bag_file(self.research_object, basename + ".provn") as provenance_file:
            self.document.serialize(provenance_file, format="provn", indent=2)
            prov_ids.append(self.provenance_ns[filename + ".provn"])

        # https://www.w3.org/Submission/prov-json/
        with write_bag_file(self.research_object, basename + ".json") as provenance_file:
            self.document.serialize(provenance_file, format="json", indent=2)
            prov_ids.append(self.provenance_ns[filename + ".json"])

        # "rdf" aka https://www.w3.org/TR/prov-o/
        # which can be serialized to ttl/nt/jsonld (and more!)

        # https://www.w3.org/TR/turtle/
        with write_bag_file(self.research_object, basename + ".ttl") as provenance_file:
            self.document.serialize(provenance_file, format="rdf", rdf_format="turtle")
            prov_ids.append(self.provenance_ns[filename + ".ttl"])

        # https://www.w3.org/TR/n-triples/
        with write_bag_file(self.research_object, basename + ".nt") as provenance_file:
            self.document.serialize(provenance_file, format="rdf", rdf_format="ntriples")
            prov_ids.append(self.provenance_ns[filename + ".nt"])

        # https://www.w3.org/TR/json-ld/
        # TODO: Use a nice JSON-LD context
        # see also https://eprints.soton.ac.uk/395985/
        # 404 Not Found on https://provenance.ecs.soton.ac.uk/prov.jsonld :(
        with write_bag_file(self.research_object, basename + ".jsonld") as provenance_file:
            self.document.serialize(provenance_file, format="rdf", rdf_format="json-ld")
            prov_ids.append(self.provenance_ns[filename + ".jsonld"])

        _logger.debug("[provenance] added provenance: %s", prov_ids)
        return prov_ids



"""Reference implementation of the CWL standards."""

__author__ = "pamstutz@veritasgenetics.com"

CWL_CONTENT_TYPES = [
    "text/plain",
    "application/json",
    "text/vnd.yaml",
    "text/yaml",
    "text/x-yaml",
    "application/x-yaml",
    "application/octet-stream",
]



from typing import Iterator, Optional

from .base import Client

def build(
    self: Client,
    recipe: Optional[str] = ...,
    image: Optional[str] = ...,
    isolated: Optional[bool] = ...,
    sandbox: Optional[bool] = ...,
    writable: Optional[bool] = ...,
    build_folder: Optional[str] = ...,
    robot_name: Optional[bool] = ...,
    ext: Optional[str] = ...,
    sudo: Optional[bool] = ...,
    stream: Optional[bool] = ...,
    force: Optional[bool] = ...,
    options: Optional[list[str]] | None = ...,
    quiet: Optional[bool] = ...,
    return_result: Optional[bool] = ...,
    sudo_options: Optional[str | list[str]] = ...,
    singularity_options: Optional[list[str]] = ...,
) -> tuple[str, Iterator[str]]: ...



from __future__ import absolute_import, division, print_function, unicode_literals

from prov.identifier import Namespace

__author__ = "Trung Dong Huynh"
__email__ = "trungdong@donggiang.com"


XSD = Namespace("xsd", "http://www.w3.org/2001/XMLSchema#")
PROV = Namespace("prov", "http://www.w3.org/ns/prov#")
XSI = Namespace("xsi", "http://www.w3.org/2001/XMLSchema-instance")

#  C1. Entities/Activities
PROV_ENTITY = PROV["Entity"]
PROV_ACTIVITY = PROV["Activity"]
PROV_GENERATION = PROV["Generation"]
PROV_USAGE = PROV["Usage"]
PROV_COMMUNICATION = PROV["Communication"]
PROV_START = PROV["Start"]
PROV_END = PROV["End"]
PROV_INVALIDATION = PROV["Invalidation"]

#  C2. Derivations
PROV_DERIVATION = PROV["Derivation"]

#  C3. Agents/Responsibility
PROV_AGENT = PROV["Agent"]
PROV_ATTRIBUTION = PROV["Attribution"]
PROV_ASSOCIATION = PROV["Association"]
PROV_DELEGATION = PROV["Delegation"]
PROV_INFLUENCE = PROV["Influence"]
#  C4. Bundles
PROV_BUNDLE = PROV["Bundle"]
#  C5. Alternate
PROV_ALTERNATE = PROV["Alternate"]
PROV_SPECIALIZATION = PROV["Specialization"]
PROV_MENTION = PROV["Mention"]
#  C6. Collections
PROV_MEMBERSHIP = PROV["Membership"]

PROV_N_MAP = {
    PROV_ENTITY: "entity",
    PROV_ACTIVITY: "activity",
    PROV_GENERATION: "wasGeneratedBy",
    PROV_USAGE: "used",
    PROV_COMMUNICATION: "wasInformedBy",
    PROV_START: "wasStartedBy",
    PROV_END: "wasEndedBy",
    PROV_INVALIDATION: "wasInvalidatedBy",
    PROV_DERIVATION: "wasDerivedFrom",
    PROV_AGENT: "agent",
    PROV_ATTRIBUTION: "wasAttributedTo",
    PROV_ASSOCIATION: "wasAssociatedWith",
    PROV_DELEGATION: "actedOnBehalfOf",
    PROV_INFLUENCE: "wasInfluencedBy",
    PROV_ALTERNATE: "alternateOf",
    PROV_SPECIALIZATION: "specializationOf",
    PROV_MENTION: "mentionOf",
    PROV_MEMBERSHIP: "hadMember",
    PROV_BUNDLE: "bundle",
}

# Records defined as subtypes in PROV-N but top level types in for example
# PROV XML also need a mapping.
ADDITIONAL_N_MAP = {
    PROV["Revision"]: "wasRevisionOf",
    PROV["Quotation"]: "wasQuotedFrom",
    PROV["PrimarySource"]: "hadPrimarySource",
    PROV["SoftwareAgent"]: "softwareAgent",
    PROV["Person"]: "person",
    PROV["Organization"]: "organization",
    PROV["Plan"]: "plan",
    PROV["Collection"]: "collection",
    PROV["EmptyCollection"]: "emptyCollection",
}

# Maps qualified names from the PROV namespace to their base class. If it
# has no baseclass it maps to itsself. This is needed for example for PROV
# XML (de)serializer where extended types are used a lot.
PROV_BASE_CLS = {
    PROV_ENTITY: PROV_ENTITY,
    PROV_ACTIVITY: PROV_ACTIVITY,
    PROV_GENERATION: PROV_GENERATION,
    PROV_USAGE: PROV_USAGE,
    PROV_COMMUNICATION: PROV_COMMUNICATION,
    PROV_START: PROV_START,
    PROV_END: PROV_END,
    PROV_INVALIDATION: PROV_INVALIDATION,
    PROV_DERIVATION: PROV_DERIVATION,
    PROV["Revision"]: PROV_DERIVATION,
    PROV["Quotation"]: PROV_DERIVATION,
    PROV["PrimarySource"]: PROV_DERIVATION,
    PROV_AGENT: PROV_AGENT,
    PROV["SoftwareAgent"]: PROV_AGENT,
    PROV["Person"]: PROV_AGENT,
    PROV["Organization"]: PROV_AGENT,
    PROV_ATTRIBUTION: PROV_ATTRIBUTION,
    PROV_ASSOCIATION: PROV_ASSOCIATION,
    PROV["Plan"]: PROV_ENTITY,
    PROV_DELEGATION: PROV_DELEGATION,
    PROV_INFLUENCE: PROV_INFLUENCE,
    PROV_ALTERNATE: PROV_ALTERNATE,
    PROV_SPECIALIZATION: PROV_SPECIALIZATION,
    PROV_MENTION: PROV_MENTION,
    PROV["Collection"]: PROV_ENTITY,
    PROV["EmptyCollection"]: PROV_ENTITY,
    PROV_MEMBERSHIP: PROV_MEMBERSHIP,
    PROV_BUNDLE: PROV_ENTITY,
}

# Identifiers for PROV's attributes
PROV_ATTR_ENTITY = PROV["entity"]
PROV_ATTR_ACTIVITY = PROV["activity"]
PROV_ATTR_TRIGGER = PROV["trigger"]
PROV_ATTR_INFORMED = PROV["informed"]
PROV_ATTR_INFORMANT = PROV["informant"]
PROV_ATTR_STARTER = PROV["starter"]
PROV_ATTR_ENDER = PROV["ender"]
PROV_ATTR_AGENT = PROV["agent"]
PROV_ATTR_PLAN = PROV["plan"]
PROV_ATTR_DELEGATE = PROV["delegate"]
PROV_ATTR_RESPONSIBLE = PROV["responsible"]
PROV_ATTR_GENERATED_ENTITY = PROV["generatedEntity"]
PROV_ATTR_USED_ENTITY = PROV["usedEntity"]
PROV_ATTR_GENERATION = PROV["generation"]
PROV_ATTR_USAGE = PROV["usage"]
PROV_ATTR_SPECIFIC_ENTITY = PROV["specificEntity"]
PROV_ATTR_GENERAL_ENTITY = PROV["generalEntity"]
PROV_ATTR_ALTERNATE1 = PROV["alternate1"]
PROV_ATTR_ALTERNATE2 = PROV["alternate2"]
PROV_ATTR_BUNDLE = PROV["bundle"]
PROV_ATTR_INFLUENCEE = PROV["influencee"]
PROV_ATTR_INFLUENCER = PROV["influencer"]
PROV_ATTR_COLLECTION = PROV["collection"]

#  Literal properties
PROV_ATTR_TIME = PROV["time"]
PROV_ATTR_STARTTIME = PROV["startTime"]
PROV_ATTR_ENDTIME = PROV["endTime"]


PROV_ATTRIBUTE_QNAMES = {
    PROV_ATTR_ENTITY,
    PROV_ATTR_ACTIVITY,
    PROV_ATTR_TRIGGER,
    PROV_ATTR_INFORMED,
    PROV_ATTR_INFORMANT,
    PROV_ATTR_STARTER,
    PROV_ATTR_ENDER,
    PROV_ATTR_AGENT,
    PROV_ATTR_PLAN,
    PROV_ATTR_DELEGATE,
    PROV_ATTR_RESPONSIBLE,
    PROV_ATTR_GENERATED_ENTITY,
    PROV_ATTR_USED_ENTITY,
    PROV_ATTR_GENERATION,
    PROV_ATTR_USAGE,
    PROV_ATTR_SPECIFIC_ENTITY,
    PROV_ATTR_GENERAL_ENTITY,
    PROV_ATTR_ALTERNATE1,
    PROV_ATTR_ALTERNATE2,
    PROV_ATTR_BUNDLE,
    PROV_ATTR_INFLUENCEE,
    PROV_ATTR_INFLUENCER,
    PROV_ATTR_COLLECTION,
}
PROV_ATTRIBUTE_LITERALS = {PROV_ATTR_TIME, PROV_ATTR_STARTTIME, PROV_ATTR_ENDTIME}

# Set of formal attributes of PROV records
PROV_ATTRIBUTES = PROV_ATTRIBUTE_QNAMES | PROV_ATTRIBUTE_LITERALS
PROV_RECORD_ATTRIBUTES = list((attr, str(attr)) for attr in PROV_ATTRIBUTES)

PROV_RECORD_IDS_MAP = dict((PROV_N_MAP[rec_type_id], rec_type_id) for rec_type_id in PROV_N_MAP)
PROV_ID_ATTRIBUTES_MAP = dict(
    (prov_id, attribute) for (prov_id, attribute) in PROV_RECORD_ATTRIBUTES
)
PROV_ATTRIBUTES_ID_MAP = dict(
    (attribute, prov_id) for (prov_id, attribute) in PROV_RECORD_ATTRIBUTES
)

# Extra definition for convenience
PROV_TYPE = PROV["type"]
PROV_LABEL = PROV["label"]
PROV_VALUE = PROV["value"]
PROV_LOCATION = PROV["location"]
PROV_ROLE = PROV["role"]

PROV_QUALIFIEDNAME = PROV["QUALIFIED_NAME"]

# XSD DATA TYPES
XSD_ANYURI = XSD["anyURI"]
XSD_QNAME = XSD["QName"]
XSD_DATETIME = XSD["dateTime"]
XSD_TIME = XSD["time"]
XSD_DATE = XSD["date"]
XSD_STRING = XSD["string"]
XSD_BOOLEAN = XSD["boolean"]
# All XSD Integer types
XSD_INTEGER = XSD["integer"]
XSD_LONG = XSD["long"]
XSD_INT = XSD["int"]
XSD_SHORT = XSD["short"]
XSD_BYTE = XSD["byte"]
XSD_NONNEGATIVEINTEGER = XSD["nonNegativeInteger"]
XSD_UNSIGNEDLONG = XSD["unsignedLong"]
XSD_UNSIGNEDINT = XSD["unsignedInt"]
XSD_UNSIGNEDSHORT = XSD["unsignedShort"]
XSD_UNSIGNEDBYTE = XSD["unsignedByte"]
XSD_POSITIVEINTEGER = XSD["positiveInteger"]
XSD_NONPOSITIVEINTEGER = XSD["nonPositiveInteger"]
XSD_NEGATIVEINTEGER = XSD["negativeInteger"]
# All XSD real number types
XSD_FLOAT = XSD["float"]
XSD_DOUBLE = XSD["double"]
XSD_DECIMAL = XSD["decimal"]



class WorkflowException(Exception):
    pass


class UnsupportedRequirement(WorkflowException):
    pass


class ArgumentException(Exception):
    """Mismatched command line arguments provided."""


class GraphTargetMissingException(WorkflowException):
    """When a ``$graph`` is encountered and there is no target and no ``main``/``#main``."""



"""Only used when there is a job script or CWLTOOL_FORCE_SHELL_POPEN=1."""

import json
import os
import subprocess  # nosec
import sys
from typing import BinaryIO, Dict, List, Optional, TextIO, Union


def handle_software_environment(cwl_env: Dict[str, str], script: str) -> Dict[str, str]:
    """Update the provided environment dict by running the script."""
    exec_env = cwl_env.copy()
    exec_env["_CWLTOOL"] = "1"
    res = subprocess.run(["bash", script], shell=False, env=exec_env)  # nosec
    if res.returncode != 0:
        sys.stderr.write("Error while using SoftwareRequirements to modify environment\n")
        return cwl_env

    env = cwl_env.copy()
    with open("output_environment.dat") as _:
        data = _.read().strip("\0")
    for line in data.split("\0"):
        key, val = line.split("=", 1)
        if key in ("_", "PWD", "SHLVL", "TMPDIR", "HOME", "_CWLTOOL"):
            # Skip some variables that are meaningful to the shell or set
            # specifically by the CWL runtime environment.
            continue
        env[key] = val
    return env


def main(argv: List[str]) -> int:
    """
    Read in the configuration JSON and execute the commands.

    The first argument is the path to the JSON dictionary file containing keys:
      "commands": an array of strings that represents the command line to run
      "cwd": A string specifying which directory to run in
      "env": a dictionary of strings containing the environment variables to set
      "stdin_path": a string (or a null) giving the path that should be piped to STDIN
      "stdout_path": a string (or a null) giving the path that should receive the STDOUT
      "stderr_path": a string (or a null) giving the path that should receive the STDERR

    The second argument is optional, it specifies a shell script to execute prior,
      and the environment variables it sets will be combined with the environment
      variables from the "env" key in the JSON dictionary from the first argument.
    """
    with open(argv[1]) as f:
        popen_description = json.load(f)
        commands = popen_description["commands"]
        cwd = popen_description["cwd"]
        env = popen_description["env"]
        env["PATH"] = os.environ.get("PATH")
        stdin_path = popen_description["stdin_path"]
        stdout_path = popen_description["stdout_path"]
        stderr_path = popen_description["stderr_path"]
        if stdin_path is not None:
            stdin: Union[BinaryIO, int] = open(stdin_path, "rb")
        else:
            stdin = subprocess.PIPE
        if stdout_path is not None:
            stdout: Union[BinaryIO, TextIO] = open(stdout_path, "wb")
        else:
            stdout = sys.stderr
        if stderr_path is not None:
            stderr: Union[BinaryIO, TextIO] = open(stderr_path, "wb")
        else:
            stderr = sys.stderr

        try:
            env_script: Optional[str] = argv[2]
        except IndexError:
            env_script = None
        if env_script is not None:
            env = handle_software_environment(env, env_script)

        sp = subprocess.Popen(  # nosec
            commands,
            shell=False,
            close_fds=True,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=env,
            cwd=cwd,
        )
        if sp.stdin:
            sp.stdin.close()
        rcode = sp.wait()
        if not isinstance(stdin, int):
            stdin.close()
        if stdout is not sys.stderr:
            stdout.close()
        if stderr is not sys.stderr:
            stderr.close()
    return rcode


if __name__ == "__main__":
    sys.exit(main(sys.argv))



from _typeshed import Incomplete
from prov.model import PROV_ACTIVITY as PROV_ACTIVITY
from prov.model import PROV_AGENT as PROV_AGENT
from prov.model import PROV_ALTERNATE as PROV_ALTERNATE
from prov.model import PROV_ASSOCIATION as PROV_ASSOCIATION
from prov.model import PROV_ATTRIBUTE_QNAMES as PROV_ATTRIBUTE_QNAMES
from prov.model import PROV_ATTRIBUTION as PROV_ATTRIBUTION
from prov.model import PROV_BUNDLE as PROV_BUNDLE
from prov.model import PROV_COMMUNICATION as PROV_COMMUNICATION
from prov.model import PROV_DELEGATION as PROV_DELEGATION
from prov.model import PROV_DERIVATION as PROV_DERIVATION
from prov.model import PROV_END as PROV_END
from prov.model import PROV_ENTITY as PROV_ENTITY
from prov.model import PROV_GENERATION as PROV_GENERATION
from prov.model import PROV_INFLUENCE as PROV_INFLUENCE
from prov.model import PROV_INVALIDATION as PROV_INVALIDATION
from prov.model import PROV_MEMBERSHIP as PROV_MEMBERSHIP
from prov.model import PROV_MENTION as PROV_MENTION
from prov.model import PROV_SPECIALIZATION as PROV_SPECIALIZATION
from prov.model import PROV_START as PROV_START
from prov.model import PROV_USAGE as PROV_USAGE
from prov.model import Identifier as Identifier
from prov.model import ProvException as ProvException
from prov.model import sorted_attributes as sorted_attributes

DOT_PROV_STYLE: Incomplete
ANNOTATION_STYLE: Incomplete
ANNOTATION_LINK_STYLE: Incomplete
ANNOTATION_START_ROW: str
ANNOTATION_ROW_TEMPLATE: str
ANNOTATION_END_ROW: str

def htlm_link_if_uri(value): ...
def prov_to_dot(
    bundle,
    show_nary: bool = ...,
    use_labels: bool = ...,
    direction: str = ...,
    show_element_attributes: bool = ...,
    show_relation_attributes: bool = ...,
): ...



"""Test singularity{,-ce} & apptainer versions."""

from subprocess import check_output  # nosec

import cwltool.singularity
from cwltool.singularity import (
    get_version,
    is_apptainer_1_or_newer,
    is_version_2_6,
    is_version_3_1_or_newer,
    is_version_3_4_or_newer,
    is_version_3_or_newer,
)


def reset_singularity_version_cache() -> None:
    """Reset the cache for testing."""
    cwltool.singularity._SINGULARITY_VERSION = None
    cwltool.singularity._SINGULARITY_FLAVOR = ""


def set_dummy_check_output(name: str, version: str) -> None:
    """Mock out subprocess.check_output."""
    cwltool.singularity.check_output = (  # type: ignore[attr-defined]
        lambda c, text: name + " version " + version  # type: ignore[assignment]
    )


def restore_check_output() -> None:
    """Undo the mock of subprocess.check_output."""
    cwltool.singularity.check_output = check_output  # type: ignore[attr-defined]


def test_get_version() -> None:
    """Confirm expected types of singularity.get_version()."""
    set_dummy_check_output("apptainer", "1.0.1")
    reset_singularity_version_cache()
    v = get_version()
    assert isinstance(v, tuple)
    assert isinstance(v[0], list)
    assert isinstance(v[1], str)
    assert cwltool.singularity._SINGULARITY_VERSION is not None  # pylint: disable=protected-access
    assert len(cwltool.singularity._SINGULARITY_FLAVOR) > 0  # pylint: disable=protected-access
    v_cached = get_version()
    assert v == v_cached

    assert v[0][0] == 1
    assert v[0][1] == 0
    assert v[0][2] == 1
    assert v[1] == "apptainer"

    set_dummy_check_output("singularity", "3.8.5")
    reset_singularity_version_cache()
    v = get_version()

    assert v[0][0] == 3
    assert v[0][1] == 8
    assert v[0][2] == 5
    assert v[1] == "singularity"
    restore_check_output()


def test_version_checks() -> None:
    """Confirm logic in the various singularity version checks."""
    set_dummy_check_output("apptainer", "1.0.1")
    reset_singularity_version_cache()
    assert is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert is_version_3_or_newer()
    assert is_version_3_1_or_newer()
    assert is_version_3_4_or_newer()

    set_dummy_check_output("apptainer", "0.0.1")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert not is_version_3_or_newer()
    assert not is_version_3_1_or_newer()
    assert not is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "0.0.1")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert not is_version_3_or_newer()
    assert not is_version_3_1_or_newer()
    assert not is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "0.1")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert not is_version_3_or_newer()
    assert not is_version_3_1_or_newer()
    assert not is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "2.6")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert is_version_2_6()
    assert not is_version_3_or_newer()
    assert not is_version_3_1_or_newer()
    assert not is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "3.0")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert is_version_3_or_newer()
    assert not is_version_3_1_or_newer()
    assert not is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "3.1")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert is_version_3_or_newer()
    assert is_version_3_1_or_newer()
    assert not is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "3.4")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert is_version_3_or_newer()
    assert is_version_3_1_or_newer()
    assert is_version_3_4_or_newer()

    set_dummy_check_output("singularity", "3.6.3")
    reset_singularity_version_cache()
    assert not is_apptainer_1_or_newer()
    assert not is_version_2_6()
    assert is_version_3_or_newer()
    assert is_version_3_1_or_newer()
    assert is_version_3_4_or_newer()
    restore_check_output()



import requests

from cwltool.main import append_word_to_default_user_agent, main


def get_user_agent() -> str:
    return requests.utils.default_headers()["User-Agent"]


def test_cwltool_in_user_agent() -> None:
    """python-requests HTTP User-Agent should include the string 'cwltool'."""
    try:
        assert main(["--version"]) == 0
    except SystemExit as err:
        assert err.code == 0
    assert "cwltool" in get_user_agent()


def test_append_word_to_default_user_agent() -> None:
    """Confirm that append_word_to_default_user_agent works."""
    word_to_append = "foobar123"
    assert word_to_append not in get_user_agent()
    append_word_to_default_user_agent(word_to_append)
    assert word_to_append in get_user_agent()



from typing import Optional

class Recipe:
    cmd: Optional[str]
    comments: list[str]
    entrypoint: Optional[str]
    environ: list[str]
    files: list[str]
    layer_files: dict[str, str]
    install: list[str]
    labels: list[str]
    ports: list[str]
    test: Optional[str]
    volumes: list[str]
    workdir: Optional[str]
    layer: int
    fromHeader: Optional[str]
    source: Optional[Recipe]
    def __init__(self, recipe: Optional[Recipe] = ..., layer: int = ...) -> None: ...



"""Tests for various command line options."""

from cwltool.utils import versionstring

from .util import get_data, get_main_output, needs_docker


def test_version() -> None:
    """Test --version."""
    return_code, stdout, stderr = get_main_output(["--version"])
    assert return_code == 0
    assert versionstring() in stdout


def test_print_supported_versions() -> None:
    """Test --print-supported-versions."""
    return_code, stdout, stderr = get_main_output(["--print-supported-versions"])
    assert return_code == 0
    assert "v1.2" in stdout


def test_empty_cmdling() -> None:
    """Test empty command line."""
    return_code, stdout, stderr = get_main_output([])
    assert return_code == 1
    assert "CWL document required, no input file was provided" in stderr


def test_tool_help() -> None:
    """Test --tool-help."""
    return_code, stdout, stderr = get_main_output(["--tool-help", get_data("tests/echo.cwl")])
    assert return_code == 0
    assert "job_order   Job input json file" in stdout


def test_basic_pack() -> None:
    """Basic test of --pack. See test_pack.py for detailed testing."""
    return_code, stdout, stderr = get_main_output(["--pack", get_data("tests/wf/revsort.cwl")])
    assert return_code == 0
    assert "$graph" in stdout


@needs_docker
def test_basic_print_subgraph() -> None:
    """Basic test of --print-subgraph. See test_subgraph.py for detailed testing."""
    return_code, stdout, stderr = get_main_output(
        [
            "--print-subgraph",
            get_data("tests/subgraph/count-lines1-wf.cwl"),
        ]
    )
    assert return_code == 0
    assert "cwlVersion" in stdout


def test_error_graph_with_no_default() -> None:
    """Ensure a useful error is printed on $graph docs that lack a main/#main."""
    exit_code, stdout, stderr = get_main_output(
        ["--tool-help", get_data("tests/wf/packed_no_main.cwl")]
    )  # could be any command except --validate
    assert exit_code == 1
    assert (
        "Tool file contains graph of multiple objects, must specify one of #echo, #cat, #collision"
        in stderr
    )


def test_skip_schemas_external_step() -> None:
    """Test that --skip-schemas works even for bad schemas in external docs."""
    exit_code, stdout, stderr = get_main_output(
        [
            "--print-rdf",
            "--skip-schemas",
            get_data("tests/wf/revsort_step_bad_schema.cwl"),
        ]
    )
    assert exit_code == 0
    assert (
        "Repeat node-elements inside property elements: " "http://www.w3.org/1999/xhtmlmeta"
    ) not in stderr
    assert (
        "Could not load extension schema https://bad.example.com/missing.ttl: "
        "Error fetching https://bad.example.com/missing.ttl"
    ) not in stderr



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class OWL(DefinedNamespace):
    allValuesFrom: URIRef
    annotatedProperty: URIRef
    annotatedSource: URIRef
    annotatedTarget: URIRef
    assertionProperty: URIRef
    cardinality: URIRef
    complementOf: URIRef
    datatypeComplementOf: URIRef
    differentFrom: URIRef
    disjointUnionOf: URIRef
    disjointWith: URIRef
    distinctMembers: URIRef
    equivalentClass: URIRef
    equivalentProperty: URIRef
    hasKey: URIRef
    hasSelf: URIRef
    hasValue: URIRef
    intersectionOf: URIRef
    inverseOf: URIRef
    maxCardinality: URIRef
    maxQualifiedCardinality: URIRef
    members: URIRef
    minCardinality: URIRef
    minQualifiedCardinality: URIRef
    onClass: URIRef
    onDataRange: URIRef
    onDatatype: URIRef
    onProperties: URIRef
    onProperty: URIRef
    oneOf: URIRef
    propertyChainAxiom: URIRef
    propertyDisjointWith: URIRef
    qualifiedCardinality: URIRef
    sameAs: URIRef
    someValuesFrom: URIRef
    sourceIndividual: URIRef
    targetIndividual: URIRef
    targetValue: URIRef
    unionOf: URIRef
    withRestrictions: URIRef
    AllDifferent: URIRef
    AllDisjointClasses: URIRef
    AllDisjointProperties: URIRef
    Annotation: URIRef
    AnnotationProperty: URIRef
    AsymmetricProperty: URIRef
    Axiom: URIRef
    Class: URIRef
    DataRange: URIRef
    DatatypeProperty: URIRef
    DeprecatedClass: URIRef
    DeprecatedProperty: URIRef
    FunctionalProperty: URIRef
    InverseFunctionalProperty: URIRef
    IrreflexiveProperty: URIRef
    NamedIndividual: URIRef
    NegativePropertyAssertion: URIRef
    ObjectProperty: URIRef
    Ontology: URIRef
    OntologyProperty: URIRef
    ReflexiveProperty: URIRef
    Restriction: URIRef
    SymmetricProperty: URIRef
    TransitiveProperty: URIRef
    backwardCompatibleWith: URIRef
    deprecated: URIRef
    incompatibleWith: URIRef
    priorVersion: URIRef
    versionInfo: URIRef
    Nothing: URIRef
    Thing: URIRef
    bottomDataProperty: URIRef
    topDataProperty: URIRef
    bottomObjectProperty: URIRef
    topObjectProperty: URIRef
    imports: URIRef
    versionIRI: URIRef



from collections.abc import Generator

from _typeshed import Incomplete
from prov.constants import PROV as PROV
from prov.constants import PROV_ACTIVITY as PROV_ACTIVITY
from prov.constants import PROV_ALTERNATE as PROV_ALTERNATE
from prov.constants import PROV_ATTR_ENDER as PROV_ATTR_ENDER
from prov.constants import PROV_ATTR_ENDTIME as PROV_ATTR_ENDTIME
from prov.constants import PROV_ATTR_INFORMANT as PROV_ATTR_INFORMANT
from prov.constants import PROV_ATTR_RESPONSIBLE as PROV_ATTR_RESPONSIBLE
from prov.constants import PROV_ATTR_STARTER as PROV_ATTR_STARTER
from prov.constants import PROV_ATTR_STARTTIME as PROV_ATTR_STARTTIME
from prov.constants import PROV_ATTR_TIME as PROV_ATTR_TIME
from prov.constants import PROV_ATTR_TRIGGER as PROV_ATTR_TRIGGER
from prov.constants import PROV_ATTR_USED_ENTITY as PROV_ATTR_USED_ENTITY
from prov.constants import PROV_BASE_CLS as PROV_BASE_CLS
from prov.constants import PROV_COMMUNICATION as PROV_COMMUNICATION
from prov.constants import PROV_DELEGATION as PROV_DELEGATION
from prov.constants import PROV_DERIVATION as PROV_DERIVATION
from prov.constants import PROV_END as PROV_END
from prov.constants import PROV_GENERATION as PROV_GENERATION
from prov.constants import PROV_ID_ATTRIBUTES_MAP as PROV_ID_ATTRIBUTES_MAP
from prov.constants import PROV_INVALIDATION as PROV_INVALIDATION
from prov.constants import PROV_LOCATION as PROV_LOCATION
from prov.constants import PROV_MENTION as PROV_MENTION
from prov.constants import PROV_N_MAP as PROV_N_MAP
from prov.constants import PROV_ROLE as PROV_ROLE
from prov.constants import PROV_START as PROV_START
from prov.constants import PROV_USAGE as PROV_USAGE
from prov.constants import XSD_QNAME as XSD_QNAME
from prov.serializers import Error as Error
from prov.serializers import Serializer as Serializer

class ProvRDFException(Error): ...

class AnonymousIDGenerator:
    def __init__(self) -> None: ...
    def get_anon_id(self, obj, local_prefix: str = ...): ...

LITERAL_XSDTYPE_MAP: Incomplete

def attr2rdf(attr): ...
def valid_qualified_name(bundle, value, xsd_qname: bool = ...): ...

class ProvRDFSerializer(Serializer):
    def serialize(
        self, stream: Incomplete | None = ..., rdf_format: str = ..., **kwargs
    ) -> None: ...
    document: Incomplete
    def deserialize(self, stream, rdf_format: str = ..., **kwargs): ...
    def valid_identifier(self, value): ...
    def encode_rdf_representation(self, value): ...
    def decode_rdf_representation(self, literal, graph): ...
    def encode_document(self, document): ...
    def encode_container(
        self,
        bundle,
        container: Incomplete | None = ...,
        identifier: Incomplete | None = ...,
    ): ...
    def decode_document(self, content, document) -> None: ...
    def decode_container(self, graph, bundle) -> None: ...

def walk(
    children, level: int = ..., path: Incomplete | None = ..., usename: bool = ...
) -> Generator[Incomplete, None, None]: ...
def literal_rdf_representation(literal): ...



"""Support for executing Docker format containers using Singularity {2,3}.x or Apptainer 1.x."""

import logging
import os
import os.path
import re
import shutil
import sys
from subprocess import check_call, check_output  # nosec
from typing import Callable, Dict, List, MutableMapping, Optional, Tuple, cast

from schema_salad.sourceline import SourceLine
from spython.main import Client
from spython.main.parse.parsers.docker import DockerParser
from spython.main.parse.writers.singularity import SingularityWriter

from .builder import Builder
from .context import RuntimeContext
from .docker import DockerCommandLineJob
from .errors import WorkflowException
from .job import ContainerCommandLineJob
from .loghandler import _logger
from .pathmapper import MapperEnt, PathMapper
from .singularity_utils import singularity_supports_userns
from .utils import CWLObjectType, create_tmp_dir, ensure_non_writable, ensure_writable

# Cached version number of singularity
# This is a list containing major and minor versions as integer.
# (The number of minor version digits can vary among different distributions,
#  therefore we need a list here.)
_SINGULARITY_VERSION: Optional[List[int]] = None
# Cached flavor / distribution of singularity
# Can be singularity, singularity-ce or apptainer
_SINGULARITY_FLAVOR: str = ""


def get_version() -> Tuple[List[int], str]:
    """
    Parse the output of 'singularity --version' to determine the flavor and version.

    Both pieces of information will be cached.

    :returns: A tuple containing:
              - A tuple with major and minor version numbers as integer.
              - A string with the name of the singularity flavor.
    """
    global _SINGULARITY_VERSION  # pylint: disable=global-statement
    global _SINGULARITY_FLAVOR  # pylint: disable=global-statement
    if _SINGULARITY_VERSION is None:
        version_output = check_output(["singularity", "--version"], text=True).strip()  # nosec

        version_match = re.match(r"(.+) version ([0-9\.]+)", version_output)
        if version_match is None:
            raise RuntimeError("Output of 'singularity --version' not recognized.")

        version_string = version_match.group(2)
        _SINGULARITY_VERSION = [int(i) for i in version_string.split(".")]
        _SINGULARITY_FLAVOR = version_match.group(1)

        _logger.debug(f"Singularity version: {version_string}" " ({_SINGULARITY_FLAVOR}.")
    return (_SINGULARITY_VERSION, _SINGULARITY_FLAVOR)


def is_apptainer_1_or_newer() -> bool:
    """
    Check if apptainer singularity distribution is version 1.0 or higher.

    Apptainer v1.0.0 is compatible with SingularityCE 3.9.5.
    See: https://github.com/apptainer/apptainer/releases
    """
    v = get_version()
    if v[1] != "apptainer":
        return False
    return v[0][0] >= 1


def is_version_2_6() -> bool:
    """
    Check if this singularity version is exactly version 2.6.

    Also returns False if the flavor is not singularity or singularity-ce.
    """
    v = get_version()
    if v[1] != "singularity" and v[1] != "singularity-ce":
        return False
    return v[0][0] == 2 and v[0][1] == 6


def is_version_3_or_newer() -> bool:
    """Check if this version is singularity version 3 or newer or equivalent."""
    if is_apptainer_1_or_newer():
        return True  # this is equivalent to singularity-ce > 3.9.5
    v = get_version()
    return v[0][0] >= 3


def is_version_3_1_or_newer() -> bool:
    """Check if this version is singularity version 3.1 or newer or equivalent."""
    if is_apptainer_1_or_newer():
        return True  # this is equivalent to singularity-ce > 3.9.5
    v = get_version()
    return v[0][0] >= 4 or (v[0][0] == 3 and v[0][1] >= 1)


def is_version_3_4_or_newer() -> bool:
    """Detect if Singularity v3.4+ is available."""
    if is_apptainer_1_or_newer():
        return True  # this is equivalent to singularity-ce > 3.9.5
    v = get_version()
    return v[0][0] >= 4 or (v[0][0] == 3 and v[0][1] >= 4)


def is_version_3_9_or_newer() -> bool:
    """Detect if Singularity v3.9+ is available."""
    if is_apptainer_1_or_newer():
        return True  # this is equivalent to singularity-ce > 3.9.5
    v = get_version()
    return v[0][0] >= 4 or (v[0][0] == 3 and v[0][1] >= 9)


def _normalize_image_id(string: str) -> str:
    return string.replace("/", "_") + ".img"


def _normalize_sif_id(string: str) -> str:
    return string.replace("/", "_") + ".sif"


class SingularityCommandLineJob(ContainerCommandLineJob):
    def __init__(
        self,
        builder: Builder,
        joborder: CWLObjectType,
        make_path_mapper: Callable[[List[CWLObjectType], str, RuntimeContext, bool], PathMapper],
        requirements: List[CWLObjectType],
        hints: List[CWLObjectType],
        name: str,
    ) -> None:
        """Builder for invoking the Singularty software container engine."""
        super().__init__(builder, joborder, make_path_mapper, requirements, hints, name)

    @staticmethod
    def get_image(
        dockerRequirement: Dict[str, str],
        pull_image: bool,
        tmp_outdir_prefix: str,
        force_pull: bool = False,
    ) -> bool:
        """
        Acquire the software container image in the specified dockerRequirement.

        Uses Singularity and returns the success as a bool. Updates the
        provided dockerRequirement with the specific dockerImageId to the full
        path of the local image, if found. Likewise the
        dockerRequirement['dockerPull'] is updated to a docker:// URI if needed.
        """
        found = False

        candidates = []

        cache_folder = None
        debug = _logger.isEnabledFor(logging.DEBUG)

        if "CWL_SINGULARITY_CACHE" in os.environ:
            cache_folder = os.environ["CWL_SINGULARITY_CACHE"]
        elif is_version_2_6() and "SINGULARITY_PULLFOLDER" in os.environ:
            cache_folder = os.environ["SINGULARITY_PULLFOLDER"]

        if "dockerFile" in dockerRequirement:
            if cache_folder is None:  # if environment variables were not set
                cache_folder = create_tmp_dir(tmp_outdir_prefix)

            absolute_path = os.path.abspath(cache_folder)
            if "dockerImageId" in dockerRequirement:
                image_name = dockerRequirement["dockerImageId"]
                image_path = os.path.join(absolute_path, image_name)
                if os.path.exists(image_path):
                    found = True
            if found is False:
                dockerfile_path = os.path.join(absolute_path, "Dockerfile")
                singularityfile_path = dockerfile_path + ".def"
                # if you do not set APPTAINER_TMPDIR will crash
                # WARNING: 'nodev' mount option set on /tmp, it could be a
                #          source of failure during build process
                # FATAL:   Unable to create build: 'noexec' mount option set on
                #          /tmp, temporary root filesystem won't be usable at this location
                with open(dockerfile_path, "w") as dfile:
                    dfile.write(dockerRequirement["dockerFile"])

                singularityfile = SingularityWriter(DockerParser(dockerfile_path).parse()).convert()
                with open(singularityfile_path, "w") as file:
                    file.write(singularityfile)

                os.environ["APPTAINER_TMPDIR"] = absolute_path
                singularity_options = ["--fakeroot"] if not shutil.which("proot") else []
                if "dockerImageId" in dockerRequirement:
                    Client.build(
                        recipe=singularityfile_path,
                        build_folder=absolute_path,
                        image=dockerRequirement["dockerImageId"],
                        sudo=False,
                        options=singularity_options,
                    )
                else:
                    Client.build(
                        recipe=singularityfile_path,
                        build_folder=absolute_path,
                        sudo=False,
                        options=singularity_options,
                    )
                found = True
        elif "dockerImageId" not in dockerRequirement and "dockerPull" in dockerRequirement:
            match = re.search(pattern=r"([a-z]*://)", string=dockerRequirement["dockerPull"])
            img_name = _normalize_image_id(dockerRequirement["dockerPull"])
            candidates.append(img_name)
            if is_version_3_or_newer():
                sif_name = _normalize_sif_id(dockerRequirement["dockerPull"])
                candidates.append(sif_name)
                dockerRequirement["dockerImageId"] = sif_name
            else:
                dockerRequirement["dockerImageId"] = img_name
            if not match:
                dockerRequirement["dockerPull"] = "docker://" + dockerRequirement["dockerPull"]
        elif "dockerImageId" in dockerRequirement:
            if os.path.isfile(dockerRequirement["dockerImageId"]):
                found = True
            candidates.append(dockerRequirement["dockerImageId"])
            candidates.append(_normalize_image_id(dockerRequirement["dockerImageId"]))
            if is_version_3_or_newer():
                candidates.append(_normalize_sif_id(dockerRequirement["dockerImageId"]))

        targets = [os.getcwd()]
        if "CWL_SINGULARITY_CACHE" in os.environ:
            targets.append(os.environ["CWL_SINGULARITY_CACHE"])
        if is_version_2_6() and "SINGULARITY_PULLFOLDER" in os.environ:
            targets.append(os.environ["SINGULARITY_PULLFOLDER"])
        for target in targets:
            for dirpath, _subdirs, files in os.walk(target):
                for entry in files:
                    if entry in candidates:
                        path = os.path.join(dirpath, entry)
                        if os.path.isfile(path):
                            _logger.info(
                                "Using local copy of Singularity image found in %s",
                                dirpath,
                            )
                            dockerRequirement["dockerImageId"] = path
                            found = True
        if (force_pull or not found) and pull_image:
            cmd = []  # type: List[str]
            if "dockerPull" in dockerRequirement:
                if cache_folder:
                    env = os.environ.copy()
                    if is_version_2_6():
                        env["SINGULARITY_PULLFOLDER"] = cache_folder
                        cmd = [
                            "singularity",
                            "pull",
                            "--force",
                            "--name",
                            dockerRequirement["dockerImageId"],
                            str(dockerRequirement["dockerPull"]),
                        ]
                    else:
                        cmd = [
                            "singularity",
                            "pull",
                            "--force",
                            "--name",
                            "{}/{}".format(cache_folder, dockerRequirement["dockerImageId"]),
                            str(dockerRequirement["dockerPull"]),
                        ]

                    _logger.info(str(cmd))
                    check_call(cmd, env=env, stdout=sys.stderr)  # nosec
                    dockerRequirement["dockerImageId"] = "{}/{}".format(
                        cache_folder, dockerRequirement["dockerImageId"]
                    )
                    found = True
                else:
                    cmd = [
                        "singularity",
                        "pull",
                        "--force",
                        "--name",
                        str(dockerRequirement["dockerImageId"]),
                        str(dockerRequirement["dockerPull"]),
                    ]
                    _logger.info(str(cmd))
                    check_call(cmd, stdout=sys.stderr)  # nosec
                    found = True

            elif "dockerLoad" in dockerRequirement:
                if is_version_3_1_or_newer():
                    if "dockerImageId" in dockerRequirement:
                        name = "{}.sif".format(dockerRequirement["dockerImageId"])
                    else:
                        name = "{}.sif".format(dockerRequirement["dockerLoad"])
                    cmd = [
                        "singularity",
                        "build",
                        name,
                        "docker-archive://{}".format(dockerRequirement["dockerLoad"]),
                    ]
                    _logger.info(str(cmd))
                    check_call(cmd, stdout=sys.stderr)  # nosec
                    found = True
                    dockerRequirement["dockerImageId"] = name
                else:
                    raise SourceLine(
                        dockerRequirement, "dockerLoad", WorkflowException, debug
                    ).makeError(
                        "dockerLoad is not currently supported when using the "
                        "Singularity runtime (version less than 3.1) for Docker containers."
                    )
            elif "dockerImport" in dockerRequirement:
                raise SourceLine(
                    dockerRequirement, "dockerImport", WorkflowException, debug
                ).makeError(
                    "dockerImport is not currently supported when using the "
                    "Singularity runtime for Docker containers."
                )

        return found

    def get_from_requirements(
        self,
        r: CWLObjectType,
        pull_image: bool,
        force_pull: bool,
        tmp_outdir_prefix: str,
    ) -> Optional[str]:
        """
        Return the filename of the Singularity image.

        (e.g. hello-world-latest.{img,sif}).
        """
        if not bool(shutil.which("singularity")):
            raise WorkflowException("singularity executable is not available")

        if not self.get_image(cast(Dict[str, str], r), pull_image, tmp_outdir_prefix, force_pull):
            raise WorkflowException("Container image {} not found".format(r["dockerImageId"]))

        if "CWL_SINGULARITY_CACHE" in os.environ:
            cache_folder = os.environ["CWL_SINGULARITY_CACHE"]
            img_path = os.path.join(cache_folder, cast(str, r["dockerImageId"]))
        else:
            img_path = cast(str, r["dockerImageId"])

        return os.path.abspath(img_path)

    @staticmethod
    def append_volume(runtime: List[str], source: str, target: str, writable: bool = False) -> None:
        """Add binding arguments to the runtime list."""
        if is_version_3_9_or_newer():
            DockerCommandLineJob.append_volume(runtime, source, target, writable, skip_mkdirs=True)
        else:
            runtime.append("--bind")
            # Mounts are writable by default, so 'rw' is optional and not
            # supported (due to a bug) in some 3.6 series releases.
            vol = f"{source}:{target}"
            if not writable:
                vol += ":ro"
            runtime.append(vol)

    def add_file_or_directory_volume(
        self, runtime: List[str], volume: MapperEnt, host_outdir_tgt: Optional[str]
    ) -> None:
        if not volume.resolved.startswith("_:"):
            if host_outdir_tgt is not None and not is_version_3_4_or_newer():
                # workaround for lack of overlapping mounts in Singularity <3.4
                if volume.type == "File":
                    os.makedirs(os.path.dirname(host_outdir_tgt), exist_ok=True)
                    shutil.copy(volume.resolved, host_outdir_tgt)
                else:
                    shutil.copytree(volume.resolved, host_outdir_tgt)
                ensure_non_writable(host_outdir_tgt)
            else:
                self.append_volume(runtime, volume.resolved, volume.target)

    def add_writable_file_volume(
        self,
        runtime: List[str],
        volume: MapperEnt,
        host_outdir_tgt: Optional[str],
        tmpdir_prefix: str,
    ) -> None:
        if host_outdir_tgt is not None and not is_version_3_4_or_newer():
            # workaround for lack of overlapping mounts in Singularity <3.4
            if self.inplace_update:
                try:
                    os.link(os.path.realpath(volume.resolved), host_outdir_tgt)
                except OSError:
                    shutil.copy(volume.resolved, host_outdir_tgt)
            else:
                shutil.copy(volume.resolved, host_outdir_tgt)
            ensure_writable(host_outdir_tgt)
        elif self.inplace_update:
            self.append_volume(runtime, volume.resolved, volume.target, writable=True)
            ensure_writable(volume.resolved)
        else:
            if host_outdir_tgt:
                # shortcut, just copy to the output directory
                # which is already going to be mounted
                if not os.path.exists(os.path.dirname(host_outdir_tgt)):
                    os.makedirs(os.path.dirname(host_outdir_tgt))
                shutil.copy(volume.resolved, host_outdir_tgt)
                ensure_writable(host_outdir_tgt)
            else:
                file_copy = os.path.join(
                    create_tmp_dir(tmpdir_prefix),
                    os.path.basename(volume.resolved),
                )
                shutil.copy(volume.resolved, file_copy)
                self.append_volume(runtime, file_copy, volume.target, writable=True)
                ensure_writable(file_copy)

    def add_writable_directory_volume(
        self,
        runtime: List[str],
        volume: MapperEnt,
        host_outdir_tgt: Optional[str],
        tmpdir_prefix: str,
    ) -> None:
        if volume.resolved.startswith("_:"):
            # Synthetic directory that needs creating first
            if not host_outdir_tgt:
                new_dir = os.path.join(
                    create_tmp_dir(tmpdir_prefix),
                    os.path.basename(volume.target),
                )
                self.append_volume(runtime, new_dir, volume.target, writable=True)
                os.makedirs(new_dir)
                # ^^ Unlike Docker, Singularity won't create directories on demand
            elif not os.path.exists(host_outdir_tgt):
                os.makedirs(host_outdir_tgt)
        else:
            if host_outdir_tgt is not None and not is_version_3_4_or_newer():
                # workaround for lack of overlapping mounts in Singularity < 3.4
                shutil.copytree(volume.resolved, host_outdir_tgt)
                ensure_writable(host_outdir_tgt)
            else:
                if self.inplace_update:
                    self.append_volume(runtime, volume.resolved, volume.target, writable=True)
                else:
                    if not host_outdir_tgt:
                        tmpdir = create_tmp_dir(tmpdir_prefix)
                        new_dir = os.path.join(tmpdir, os.path.basename(volume.resolved))
                        shutil.copytree(volume.resolved, new_dir)
                        self.append_volume(runtime, new_dir, volume.target, writable=True)
                    else:
                        shutil.copytree(volume.resolved, host_outdir_tgt)
                    ensure_writable(host_outdir_tgt or new_dir)

    def _required_env(self) -> Dict[str, str]:
        return {
            "TMPDIR": self.CONTAINER_TMPDIR,
            "HOME": self.builder.outdir,
        }

    def create_runtime(
        self, env: MutableMapping[str, str], runtime_context: RuntimeContext
    ) -> Tuple[List[str], Optional[str]]:
        """Return the Singularity runtime list of commands and options."""
        any_path_okay = self.builder.get_requirement("DockerRequirement")[1] or False
        runtime = [
            "singularity",
            "--quiet",
            "exec",
            "--contain",
            "--ipc",
            "--cleanenv",
        ]
        if singularity_supports_userns():
            runtime.append("--userns")
        else:
            runtime.append("--pid")

        container_HOME: Optional[str] = None
        if is_version_3_1_or_newer():
            # Remove HOME, as passed in a special way (restore it below)
            container_HOME = self.environment.pop("HOME")
            runtime.append("--home")
            runtime.append(
                "{}:{}".format(
                    os.path.realpath(self.outdir),
                    container_HOME,
                )
            )
        else:
            self.append_volume(
                runtime,
                os.path.realpath(self.outdir),
                self.environment["HOME"],
                writable=True,
            )

        self.append_volume(
            runtime, os.path.realpath(self.tmpdir), self.CONTAINER_TMPDIR, writable=True
        )

        self.add_volumes(
            self.pathmapper,
            runtime,
            any_path_okay=True,
            secret_store=runtime_context.secret_store,
            tmpdir_prefix=runtime_context.tmpdir_prefix,
        )
        if self.generatemapper is not None:
            self.add_volumes(
                self.generatemapper,
                runtime,
                any_path_okay=any_path_okay,
                secret_store=runtime_context.secret_store,
                tmpdir_prefix=runtime_context.tmpdir_prefix,
            )

        runtime.append("--pwd")
        runtime.append(self.builder.outdir)

        if self.networkaccess:
            if runtime_context.custom_net:
                runtime.extend(["--net", "--network", runtime_context.custom_net])
        else:
            runtime.extend(["--net", "--network", "none"])

        if self.builder.resources.get("cudaDeviceCount"):
            runtime.append("--nv")

        for name, value in self.environment.items():
            env[f"SINGULARITYENV_{name}"] = str(value)

        if container_HOME:
            # Restore HOME if we removed it above.
            self.environment["HOME"] = container_HOME
        return (runtime, None)



import json
import os
import tempfile
from collections.abc import Sized
from functools import partial
from io import StringIO
from pathlib import Path
from typing import Dict

import pytest
from schema_salad.utils import yaml_no_ts

import cwltool.pack
import cwltool.workflow
from cwltool.context import LoadingContext
from cwltool.load_tool import fetch_document, resolve_and_validate_document
from cwltool.main import main, make_relative, print_pack
from cwltool.resolver import tool_resolver
from cwltool.utils import adjustDirObjs, adjustFileObjs

from .util import get_data, needs_docker


@pytest.mark.parametrize(
    "unpacked,expected",
    [
        ("tests/wf/revsort.cwl", "tests/wf/expect_packed.cwl"),
        (
            "tests/wf/operation/operation-single.cwl",
            "tests/wf/operation/expect_operation-single_packed.cwl",
        ),
        ("tests/wf/trick_revsort.cwl", "tests/wf/expect_trick_packed.cwl"),
        (
            "tests/wf/iwd-passthrough1.cwl",
            "tests/wf/expect_iwd-passthrough1_packed.cwl",
        ),
        (
            "tests/wf/revsort_datetime.cwl",
            "tests/wf/expect_revsort_datetime_packed.cwl",
        ),
    ],
)
def test_packing(unpacked: str, expected: str) -> None:
    """Compare expected version reality with various workflows and --pack."""
    loadingContext, workflowobj, uri = fetch_document(get_data(unpacked))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri)

    packed = json.loads(print_pack(loadingContext, uri))
    context_dir = os.path.abspath(os.path.dirname(get_data(unpacked)))
    adjustFileObjs(packed, partial(make_relative, context_dir))
    adjustDirObjs(packed, partial(make_relative, context_dir))

    with open(get_data(expected)) as packed_file:
        expect_packed = json.load(packed_file)

    if "$schemas" in expect_packed:
        assert "$schemas" in packed
        packed_schemas = packed["$schemas"]
        assert isinstance(packed_schemas, Sized)
        assert len(packed_schemas) == len(expect_packed["$schemas"])
        del packed["$schemas"]
        del expect_packed["$schemas"]

    assert packed == expect_packed


def test_pack_single_tool() -> None:
    loadingContext, workflowobj, uri = fetch_document(get_data("tests/wf/formattest.cwl"))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri)
    loader = loadingContext.loader
    assert loader
    loader.resolve_ref(uri)[0]

    packed = cwltool.pack.pack(loadingContext, uri)
    assert "$schemas" in packed


def test_pack_fragment() -> None:
    yaml = yaml_no_ts()

    with open(get_data("tests/wf/scatter2_subwf.cwl")) as packed_file:
        expect_packed = yaml.load(packed_file)

    loadingContext, workflowobj, uri = fetch_document(get_data("tests/wf/scatter2.cwl"))
    packed = cwltool.pack.pack(loadingContext, uri + "#scatterstep/mysub")
    adjustFileObjs(packed, partial(make_relative, os.path.abspath(get_data("tests/wf"))))
    adjustDirObjs(packed, partial(make_relative, os.path.abspath(get_data("tests/wf"))))

    packed_result = json.dumps(packed, sort_keys=True, indent=2)
    expected = json.dumps(expect_packed, sort_keys=True, indent=2)

    assert packed_result == expected


def test_pack_rewrites() -> None:
    rewrites: Dict[str, str] = {}

    loadingContext, workflowobj, uri = fetch_document(get_data("tests/wf/default-wf5.cwl"))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri)
    loader = loadingContext.loader
    assert loader
    loader.resolve_ref(uri)[0]

    cwltool.pack.pack(
        loadingContext,
        uri,
        rewrite_out=rewrites,
    )

    assert len(rewrites) == 6


cwl_missing_version_paths = [
    "tests/wf/hello_single_tool.cwl",
    "tests/wf/hello-workflow.cwl",
]


@pytest.mark.parametrize("cwl_path", cwl_missing_version_paths)
def test_pack_missing_cwlVersion(cwl_path: str) -> None:
    """Test generated pack output for `cwlVersion` in case of 1 tool workflow and 1 step workflow."""
    # Testing single tool workflow
    loadingContext, workflowobj, uri = fetch_document(get_data(cwl_path))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri)
    loader = loadingContext.loader
    assert loader
    loader.resolve_ref(uri)[0]

    # generate pack output dict
    packed = json.loads(print_pack(loadingContext, uri))

    assert packed["cwlVersion"] == "v1.0"


def test_pack_idempotence_tool(tmp_path: Path) -> None:
    """Ensure that pack produces exactly the same document for an already packed CommandLineTool."""
    _pack_idempotently("tests/wf/hello_single_tool.cwl", tmp_path)


def test_pack_idempotence_workflow(tmp_path: Path) -> None:
    """Ensure that pack produces exactly the same document for an already packed workflow."""
    _pack_idempotently("tests/wf/count-lines1-wf.cwl", tmp_path)


def _pack_idempotently(document: str, tmp_path: Path) -> None:
    loadingContext, workflowobj, uri = fetch_document(get_data(document))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri)
    loader = loadingContext.loader
    assert loader
    loader.resolve_ref(uri)[0]

    # generate pack output dict
    packed_text = print_pack(loadingContext, uri)
    packed = json.loads(packed_text)

    tmp_name = tmp_path / "packed.cwl"
    tmp = tmp_name.open(mode="w")
    tmp.write(packed_text)
    tmp.flush()
    tmp.close()

    loadingContext, workflowobj, uri2 = fetch_document(tmp.name)
    loadingContext.do_update = False
    loadingContext, uri2 = resolve_and_validate_document(loadingContext, workflowobj, uri2)
    loader2 = loadingContext.loader
    assert loader2
    loader2.resolve_ref(uri2)[0]

    # generate pack output dict
    packed_text = print_pack(loadingContext, uri2)
    double_packed = json.loads(packed_text)

    assert uri != uri2
    assert packed == double_packed


cwl_to_run = [
    ("tests/wf/count-lines1-wf.cwl", "tests/wf/wc-job.json", False),
    ("tests/wf/formattest.cwl", "tests/wf/formattest-job.json", True),
]


@needs_docker
@pytest.mark.parametrize("wf_path,job_path,namespaced", cwl_to_run)
def test_packed_workflow_execution(
    wf_path: str, job_path: str, namespaced: bool, tmp_path: Path
) -> None:
    loadingContext = LoadingContext()
    loadingContext.resolver = tool_resolver
    loadingContext, workflowobj, uri = fetch_document(get_data(wf_path), loadingContext)
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri)
    loader = loadingContext.loader
    assert loader
    loader.resolve_ref(uri)[0]
    packed = json.loads(print_pack(loadingContext, uri))

    assert not namespaced or "$namespaces" in packed

    wf_packed_handle, wf_packed_path = tempfile.mkstemp()
    with open(wf_packed_path, "w") as temp_file:
        json.dump(packed, temp_file)

    normal_output = StringIO()
    packed_output = StringIO()

    normal_params = ["--outdir", str(tmp_path), get_data(wf_path), get_data(job_path)]
    packed_params = [
        "--outdir",
        str(tmp_path),
        "--debug",
        wf_packed_path,
        get_data(job_path),
    ]

    assert main(normal_params, stdout=normal_output) == 0
    assert main(packed_params, stdout=packed_output) == 0

    assert json.loads(packed_output.getvalue()) == json.loads(normal_output.getvalue())

    os.close(wf_packed_handle)
    os.remove(wf_packed_path)



"""Tests for docker engine."""

import json
import re
from pathlib import Path
from shutil import which

import pytest

from cwltool.main import main

from .util import (
    get_data,
    get_main_output,
    needs_docker,
    needs_podman,
    needs_singularity,
)


@needs_docker
def test_docker_workflow(tmp_path: Path) -> None:
    """Basic test for docker with a CWL Workflow."""
    result_code, _, stderr = get_main_output(
        [
            "--default-container",
            "docker.io/debian:stable-slim",
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/hello-workflow.cwl"),
            "--usermessage",
            "hello",
        ]
    )
    assert "completed success" in stderr
    assert (tmp_path / "response.txt").read_text("utf-8") == "hello"
    assert result_code == 0


def test_docker_iwdr() -> None:
    result_code = main(
        [
            "--default-container",
            "docker.io/debian:stable-slim",
            get_data("tests/wf/iwdr-entry.cwl"),
            "--message",
            "hello",
        ]
    )
    docker_installed = bool(which("docker"))
    if docker_installed:
        assert result_code == 0
    else:
        assert result_code != 0


@needs_docker
def test_docker_incorrect_image_pull() -> None:
    result_code = main(
        [
            "--default-container",
            "non-existant-weird-image",
            get_data("tests/wf/hello-workflow.cwl"),
            "--usermessage",
            "hello",
        ]
    )
    assert result_code != 0


@needs_docker
def test_docker_file_mount() -> None:
    # test for bug in
    # ContainerCommandLineJob.create_file_and_add_volume()
    #
    # the bug was that it would use the file literal contents as the
    # temporary file name, which can easily result in a file name that
    # is too long or otherwise invalid.  This test case uses ".."
    result_code = main(
        [get_data("tests/wf/literalfile.cwl"), get_data("tests/wf/literalfile-job.yml")]
    )
    assert result_code == 0


@needs_docker
def test_docker_strict_cpu_limit(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--strict-cpu-limit",
            "--default-container",
            "docker.io/debian:stable-slim",
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/cores_float.cwl"),
        ]
    )
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert result_code == 0
    assert "--cpus=2" in stderr


@needs_docker
def test_docker_strict_memory_limit(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--strict-memory-limit",
            "--default-container",
            "docker.io/debian:stable-slim",
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/storage_float.cwl"),
        ]
    )
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert result_code == 0
    assert "--memory=255m" in stderr


@needs_docker
def test_docker_strict_cpu_limit_warning(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--default-container",
            "docker.io/debian:stable-slim",
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/cores_float.cwl"),
        ]
    )
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert result_code == 0
    assert "Skipping Docker software container '--cpus' limit" in stderr


@needs_docker
def test_docker_strict_memory_limit_warning(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--default-container",
            "docker.io/debian:stable-slim",
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/storage_float.cwl"),
        ]
    )
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert result_code == 0
    assert "Skipping Docker software container '--memory' limit" in stderr


@needs_docker
def test_docker_required_secfile(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--outdir",
            str(tmp_path),
            get_data("tests/secondary-files-required-container.cwl"),
        ]
    )
    assert result_code == 0, stderr
    assert (
        json.loads(stdout)["output"]["secondaryFiles"][0]["checksum"]
        == "sha1$da39a3ee5e6b4b0d3255bfef95601890afd80709"
    )


@needs_podman
def test_podman_required_secfile(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--podman",
            "--outdir",
            str(tmp_path),
            get_data("tests/secondary-files-required-container.cwl"),
        ]
    )
    assert result_code == 0, stderr
    assert (
        json.loads(stdout)["output"]["secondaryFiles"][0]["checksum"]
        == "sha1$da39a3ee5e6b4b0d3255bfef95601890afd80709"
    )


@needs_singularity
def test_singularity_required_secfile(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    singularity_dir = tmp_path / "singularity"
    singularity_dir.mkdir()
    monkeypatch.setenv("CWL_SINGULARITY_CACHE", str(singularity_dir))

    result_code, stdout, stderr = get_main_output(
        [
            "--singularity",
            "--outdir",
            str(tmp_path / "out"),
            get_data("tests/secondary-files-required-container.cwl"),
        ]
    )
    assert result_code == 0, stderr
    assert (
        json.loads(stdout)["output"]["secondaryFiles"][0]["checksum"]
        == "sha1$da39a3ee5e6b4b0d3255bfef95601890afd80709"
    )


@needs_docker
def test_docker_required_missing_secfile(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--outdir",
            str(tmp_path),
            get_data("tests/secondary-files-required-missing-container.cwl"),
        ]
    )
    assert result_code == 1, stderr
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "Job error:" in stderr
    assert "Error collecting output for parameter 'output'" in stderr
    assert (
        "tests/secondary-files-required-missing-container.cwl:16:5: Missing required secondary file"
    )
    assert "file.ext3" in stderr


@needs_podman
def test_podman_required_missing_secfile(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--podman",
            "--outdir",
            str(tmp_path),
            get_data("tests/secondary-files-required-missing-container.cwl"),
        ]
    )
    assert result_code == 1, stderr
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "Job error:" in stderr
    assert "Error collecting output for parameter 'output'" in stderr
    assert (
        "tests/secondary-files-required-missing-container.cwl:16:5: Missing required secondary file"
    )
    assert "file.ext3" in stderr


@needs_singularity
def test_singularity_required_missing_secfile(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    singularity_dir = tmp_path / "singularity"
    singularity_dir.mkdir()
    monkeypatch.setenv("CWL_SINGULARITY_CACHE", str(singularity_dir))
    result_code, stdout, stderr = get_main_output(
        [
            "--singularity",
            "--outdir",
            str(tmp_path),
            get_data("tests/secondary-files-required-missing-container.cwl"),
        ]
    )
    assert result_code == 1, stderr
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "Job error:" in stderr
    assert "Error collecting output for parameter 'output'" in stderr
    assert (
        "tests/secondary-files-required-missing-container.cwl:16:5: Missing required secondary file"
    )
    assert "file.ext3" in stderr


@needs_docker
def test_docker_shm_size(tmp_path: Path) -> None:
    result_code, stdout, stderr = get_main_output(
        [
            "--enable-ext",
            "--default-container",
            "docker.io/debian:stable-slim",
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/shm_size.cwl"),
        ]
    )
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert result_code == 0
    assert "--shm-size=128m" in stderr



from typing import Optional

from ..recipe import Recipe
from .base import WriterBase as WriterBase

class SingularityWriter(WriterBase):
    name: str
    def __init__(self, recipe: Optional[dict[str, Recipe]] = ...) -> None: ...
    def validate(self) -> None: ...
    def convert(self, runscript: str = ..., force: bool = ...) -> str: ...



"""Test functions."""

import atexit
import contextlib
import io
import json
import os
import shutil
import subprocess
import sys
from contextlib import ExitStack
from pathlib import Path
from typing import Dict, Generator, List, Mapping, Optional, Tuple, Union

import pytest

from cwltool.env_to_stdout import deserialize_env
from cwltool.main import main
from cwltool.singularity import is_version_2_6, is_version_3_or_newer
from cwltool.utils import as_file, files


def force_default_container(default_container_id: str, _: str) -> str:
    return default_container_id


def get_data(filename: str) -> str:
    # normalizing path depending on OS or else it will cause problem when joining path
    filename = os.path.normpath(filename)
    filepath = None
    try:
        file_manager = ExitStack()
        atexit.register(file_manager.close)
        traversable = files("cwltool") / filename
        filepath = file_manager.enter_context(as_file(traversable))
    except ModuleNotFoundError:
        pass
    if not filepath or not os.path.isfile(filepath):
        filepath = Path(os.path.dirname(__file__)) / ".." / filename
    return str(filepath.resolve())


needs_docker = pytest.mark.skipif(
    not bool(shutil.which("docker")),
    reason="Requires the docker executable on the system path.",
)

needs_singularity = pytest.mark.skipif(
    not bool(shutil.which("singularity")),
    reason="Requires the singularity executable on the system path.",
)

needs_singularity_2_6 = pytest.mark.skipif(
    not bool(shutil.which("singularity") and is_version_2_6()),
    reason="Requires that version 2.6.x of singularity executable version is on the system path.",
)

needs_singularity_3_or_newer = pytest.mark.skipif(
    (not bool(shutil.which("singularity"))) or (not is_version_3_or_newer()),
    reason="Requires that version 3.x of singularity executable version is on the system path.",
)

needs_podman = pytest.mark.skipif(
    not bool(shutil.which("podman")),
    reason="Requires the podman executable on the system path.",
)

_env_accepts_null: Optional[bool] = None


def env_accepts_null() -> bool:
    """Return True iff the env command on this host accepts `-0`."""
    global _env_accepts_null
    if _env_accepts_null is None:
        result = subprocess.run(
            ["env", "-0"],
            capture_output=True,
            encoding="utf-8",
        )
        _env_accepts_null = result.returncode == 0

    return _env_accepts_null


def get_main_output(
    args: List[str],
    replacement_env: Optional[Mapping[str, str]] = None,
    extra_env: Optional[Mapping[str, str]] = None,
    monkeypatch: Optional[pytest.MonkeyPatch] = None,
) -> Tuple[Optional[int], str, str]:
    """Run cwltool main.

    args: the command line args to call it with

    replacement_env: a total replacement of the environment

    extra_env: add these to the environment used

    monkeypatch: required if changing the environment

    Returns (return code, stdout, stderr)
    """
    stdout = io.StringIO()
    stderr = io.StringIO()
    if replacement_env is not None:
        assert monkeypatch is not None
        monkeypatch.setattr(os, "environ", replacement_env)

    if extra_env is not None:
        assert monkeypatch is not None
        for k, v in extra_env.items():
            monkeypatch.setenv(k, v)

    try:
        rc = main(argsl=args, stdout=stdout, stderr=stderr)
    except SystemExit as e:
        if isinstance(e.code, int):
            rc = e.code
        else:
            rc = sys.maxsize
    return (
        rc,
        stdout.getvalue(),
        stderr.getvalue(),
    )


def get_tool_env(
    tmp_path: Path,
    flag_args: List[str],
    inputs_file: Optional[str] = None,
    replacement_env: Optional[Mapping[str, str]] = None,
    extra_env: Optional[Mapping[str, str]] = None,
    monkeypatch: Optional[pytest.MonkeyPatch] = None,
    runtime_env_accepts_null: Optional[bool] = None,
) -> Dict[str, str]:
    """Get the env vars for a tool's invocation."""
    # GNU env accepts the -0 option to end each variable's
    # printing with "\0". No such luck on BSD-ish.
    #
    # runtime_env_accepts_null is None => figure it out, otherwise
    # use wrapped bool (because containers).
    if runtime_env_accepts_null is None:
        runtime_env_accepts_null = env_accepts_null()

    args = flag_args.copy()
    if runtime_env_accepts_null:
        args.append(get_data("tests/env3.cwl"))
    else:
        args.append(get_data("tests/env4.cwl"))

    if inputs_file is not None:
        args.append(inputs_file)

    with working_directory(tmp_path):
        rc, stdout, stderr = get_main_output(
            args,
            replacement_env=replacement_env,
            extra_env=extra_env,
            monkeypatch=monkeypatch,
        )
        assert rc == 0, stdout + "\n" + stderr

        output = json.loads(stdout)
        with open(output["env"]["path"]) as _:
            return deserialize_env(_.read())


@contextlib.contextmanager
def working_directory(path: Union[str, Path]) -> Generator[None, None, None]:
    """Change working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)



import hashlib
import os
import uuid

from prov.identifier import Namespace

__citation__ = "https://doi.org/10.5281/zenodo.1208477"

# NOTE: Semantic versioning of the CWLProv Research Object
# **and** the cwlprov files
#
# Rough guide (major.minor.patch):
# 1. Bump major number if removing/"breaking" resources or PROV statements
# 2. Bump minor number if adding resources or PROV statements
# 3. Bump patch number for non-breaking non-adding changes,
#    e.g. fixing broken relative paths
CWLPROV_VERSION = "https://w3id.org/cwl/prov/0.6.0"

# Research Object folders
METADATA = "metadata"
DATA = "data"
WORKFLOW = "workflow"
SNAPSHOT = "snapshot"
# sub-folders
MAIN = os.path.join(WORKFLOW, "main")
PROVENANCE = os.path.join(METADATA, "provenance")
LOGS = os.path.join(METADATA, "logs")
WFDESC = Namespace("wfdesc", "http://purl.org/wf4ever/wfdesc#")
WFPROV = Namespace("wfprov", "http://purl.org/wf4ever/wfprov#")
WF4EVER = Namespace("wf4ever", "http://purl.org/wf4ever/wf4ever#")
RO = Namespace("ro", "http://purl.org/wf4ever/ro#")
ORE = Namespace("ore", "http://www.openarchives.org/ore/terms/")
FOAF = Namespace("foaf", "http://xmlns.com/foaf/0.1/")
SCHEMA = Namespace("schema", "http://schema.org/")
CWLPROV = Namespace("cwlprov", "https://w3id.org/cwl/prov#")
ORCID = Namespace("orcid", "https://orcid.org/")
UUID = Namespace("id", "urn:uuid:")

# BagIt and YAML always use UTF-8
ENCODING = "UTF-8"
TEXT_PLAIN = f"text/plain; charset={ENCODING!r}"

# sha1, compatible with the File type's "checksum" field
# e.g. "checksum" = "sha1$47a013e660d408619d894b20806b1d5086aab03b"
# See ./cwltool/schemas/v1.0/Process.yml
Hasher = hashlib.sha1
SHA1 = "sha1"
SHA256 = "sha256"
SHA512 = "sha512"

# TODO: Better identifiers for user, at least
# these should be preserved in ~/.config/cwl for every execution
# on this host
USER_UUID = uuid.uuid4().urn
ACCOUNT_UUID = uuid.uuid4().urn



from datetime import datetime
from typing import IO, Any, Dict, Iterable, List, Set, Tuple

from _typeshed import Incomplete
from prov.constants import *

# from prov import Error as Error, serializers as serializers
from prov.identifier import Identifier, Namespace, QualifiedName

logger: Incomplete

# def parse_xsd_datetime(value): ...

DATATYPE_PARSERS: Incomplete
XSD_DATATYPE_PARSERS: Incomplete

# def first(a_set): ...

_attributes_type = Dict[str | Identifier, Any] | List[Tuple[str | Identifier, Any]]

class ProvRecord:
    FORMAL_ATTRIBUTES: Incomplete
    def __init__(
        self,
        bundle: str,
        identifier: Identifier | str,
        attributes: Dict[str, str] | None = ...,
    ) -> None: ...
    def __hash__(self) -> int: ...
    def copy(self) -> ProvRecord: ...
    def get_type(self) -> str: ...
    def get_asserted_types(self) -> Set[str]: ...
    def add_asserted_type(self, type_identifier: str | QualifiedName) -> None: ...
    def get_attribute(self, attr_name: str) -> Set[str]: ...
    @property
    def identifier(self) -> Identifier: ...
    @property
    def attributes(self) -> List[Tuple[str, str]]: ...
    @property
    def args(self) -> Tuple[str, ...]: ...
    @property
    def formal_attributes(self) -> Tuple[Tuple[str, str], ...]: ...
    @property
    def extra_attributes(self) -> Tuple[Tuple[str, str], ...]: ...
    @property
    def bundle(self) -> "ProvBundle": ...
    @property
    def label(self) -> str: ...
    @property
    def value(self) -> str: ...
    def add_attributes(
        self,
        attributes: _attributes_type,
    ) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def get_provn(self) -> str: ...
    def is_element(self) -> bool: ...
    def is_relation(self) -> bool: ...

class ProvElement(ProvRecord):
    def __init__(
        self,
        bundle: "ProvBundle",
        identifier: str,
        attributes: _attributes_type,
    ) -> None: ...
    def is_element(self) -> bool: ...

class ProvRelation(ProvRecord):
    def is_relation(self) -> bool: ...

class ProvEntity(ProvElement):
    def wasGeneratedBy(
        self,
        activity: str,
        time: datetime | str | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvGeneration: ...
    def wasInvalidatedBy(
        self,
        activity: str,
        time: datetime | str | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvInvalidation: ...
    def wasDerivedFrom(
        self,
        usedEntity: str,
        activity: str | None = ...,
        generation: Incomplete | None = ...,
        usage: Incomplete | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvDerivation: ...
    def wasAttributedTo(
        self,
        agent: str,
        attributes: _attributes_type | None = None,
    ) -> ProvAttribution: ...
    def alternateOf(self, alternate2: str) -> ProvElement: ...
    def specializationOf(self, generalEntity: str) -> ProvSpecialization: ...
    def hadMember(self, entity: str) -> ProvMembership: ...

class ProvActivity(ProvElement):
    FORMAL_ATTRIBUTES: Incomplete
    def set_time(
        self, startTime: Incomplete | None = ..., endTime: Incomplete | None = ...
    ) -> None: ...
    def get_startTime(self) -> datetime: ...
    def get_endTime(self) -> datetime: ...
    def used(
        self,
        entity: str,
        time: datetime | str | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvActivity: ...
    def wasInformedBy(
        self,
        informant: str,
        attributes: _attributes_type | None = None,
    ) -> ProvActivity: ...
    def wasStartedBy(
        self,
        trigger: str,
        starter: str | None = ...,
        time: datetime | str | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvActivity: ...
    def wasEndedBy(
        self,
        trigger: str,
        ender: str | None = ...,
        time: datetime | str | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvActivity: ...
    def wasAssociatedWith(
        self,
        agent: str,
        plan: str | None = ...,
        attributes: _attributes_type | None = None,
    ) -> ProvActivity: ...

class ProvGeneration(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvUsage(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvCommunication(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvStart(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvEnd(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvInvalidation(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvDerivation(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvAgent(ProvElement):
    def actedOnBehalfOf(
        self,
        responsible: str,
        activity: str | None = ...,
        attributes: _attributes_type | None = ...,
    ) -> ProvAgent: ...

class ProvAttribution(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvAssociation(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvDelegation(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvInfluence(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvSpecialization(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvAlternate(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

class ProvMention(ProvSpecialization):
    FORMAL_ATTRIBUTES: Incomplete

class ProvMembership(ProvRelation):
    FORMAL_ATTRIBUTES: Incomplete

PROV_REC_CLS: Incomplete
DEFAULT_NAMESPACES: Incomplete

class ProvBundle:
    def __init__(
        self,
        records: Iterable[ProvRecord] | None = ...,
        identifier: str | None = ...,
        namespaces: Iterable[Namespace] | None = ...,
        document: ProvDocument | None = ...,
    ) -> None: ...
    @property
    def namespaces(self) -> Set[Namespace]: ...
    @property
    def default_ns_uri(self) -> str | None: ...
    @property
    def document(self) -> ProvDocument | None: ...
    @property
    def identifier(self) -> str | None | QualifiedName: ...
    @property
    def records(self) -> List[ProvRecord]: ...
    def set_default_namespace(self, uri: Namespace) -> None: ...
    def get_default_namespace(self) -> Namespace: ...
    def add_namespace(
        self, namespace_or_prefix: Namespace | str, uri: str | None = ...
    ) -> Namespace: ...
    def get_registered_namespaces(self) -> Iterable[Namespace]: ...
    def valid_qualified_name(
        self, identifier: QualifiedName | Tuple[str, Identifier]
    ) -> QualifiedName | None: ...
    def get_records(
        self,
        class_or_type_or_tuple: (
            type | type[int | str] | Tuple[type | type[int | str] | Tuple[Any, ...], ...] | None
        ) = ...,
    ) -> List[ProvRecord]: ...
    def get_record(self, identifier: Identifier | None) -> ProvRecord | List[ProvRecord] | None: ...
    def is_document(self) -> bool: ...
    def is_bundle(self) -> bool: ...
    def has_bundles(self) -> bool: ...
    @property
    def bundles(self) -> Iterable[ProvBundle]: ...
    def get_provn(self, _indent_level: int = ...) -> str: ...
    def unified(self) -> ProvBundle: ...
    def update(self, other: ProvBundle) -> None: ...
    def new_record(
        self,
        record_type: PROV_REC_CLS,
        identifier: str,
        attributes: _attributes_type | None = None,
        other_attributes: _attributes_type | None = None,
    ) -> ProvRecord: ...
    def add_record(self, record: ProvRecord) -> ProvRecord: ...
    def entity(
        self,
        identifier: str | QualifiedName,
        other_attributes: _attributes_type | None = None,
    ) -> ProvEntity: ...
    def activity(
        self,
        identifier: str | QualifiedName,
        startTime: datetime | str | None = ...,
        endTime: datetime | str | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvActivity: ...
    def generation(
        self,
        entity: ProvEntity | str,
        activity: ProvActivity | str | None = ...,
        time: datetime | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvGeneration: ...
    def usage(
        self,
        activity: ProvActivity | str,
        entity: ProvEntity | str | None = ...,
        time: datetime | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvUsage: ...
    def start(
        self,
        activity: ProvActivity | ProvAgent | str,
        trigger: ProvEntity | None = ...,
        starter: ProvActivity | ProvAgent | str | None = ...,
        time: datetime | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvStart: ...
    def end(
        self,
        activity: ProvActivity | str,
        trigger: ProvEntity | None = ...,
        ender: ProvActivity | str | None = ...,
        time: datetime | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvEnd: ...
    def invalidation(
        self,
        entity: ProvEntity | str,
        activity: ProvActivity | str | None = ...,
        time: datetime | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvInvalidation: ...
    def communication(
        self,
        informed: ProvActivity | str,
        informant: ProvActivity | str,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvCommunication: ...
    def agent(
        self,
        identifier: Identifier | str,
        other_attributes: _attributes_type | None = None,
    ) -> ProvAgent: ...
    def attribution(
        self,
        entity: ProvEntity | str,
        agent: ProvAgent | str,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvAttribution: ...
    def association(
        self,
        activity: ProvActivity | str,
        agent: ProvAgent | str | None = ...,
        plan: ProvEntity | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvAssociation: ...
    def delegation(
        self,
        delegate: ProvAgent | str,
        responsible: ProvAgent | str,
        activity: ProvActivity | str | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvDelegation: ...
    def influence(
        self,
        influencee: ProvEntity | ProvActivity | ProvAgent | str,
        influencer: ProvEntity | ProvActivity | ProvAgent | str,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvInfluence: ...
    def derivation(
        self,
        generatedEntity: ProvEntity | str,
        usedEntity: ProvEntity | str,
        activity: ProvActivity | str | None = ...,
        generation: ProvActivity | str | None = ...,
        usage: Incomplete | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvDerivation: ...
    def revision(
        self,
        generatedEntity: ProvEntity | str,
        usedEntity: ProvEntity | str,
        activity: ProvActivity | str | None = ...,
        generation: ProvActivity | str | None = ...,
        usage: Incomplete | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvRecord: ...
    def quotation(
        self,
        generatedEntity: ProvEntity | str,
        usedEntity: ProvEntity | str,
        activity: ProvActivity | str | None = ...,
        generation: ProvActivity | str | None = ...,
        usage: Incomplete | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvRecord: ...
    def primary_source(
        self,
        generatedEntity: ProvEntity | str,
        usedEntity: ProvEntity | str,
        activity: ProvActivity | str | None = ...,
        generation: ProvActivity | str | None = ...,
        usage: Incomplete | None = ...,
        identifier: Identifier | None = ...,
        other_attributes: _attributes_type | None = None,
    ) -> ProvRecord: ...
    def specialization(
        self, specificEntity: ProvEntity | str, generalEntity: ProvEntity | str
    ) -> ProvRecord: ...
    def alternate(
        self, alternate1: ProvEntity | str, alternate2: ProvEntity | str
    ) -> ProvRecord: ...
    def mention(
        self,
        specificEntity: ProvEntity | str,
        generalEntity: ProvEntity | str,
        bundle: Incomplete,
    ) -> ProvRecord: ...
    def collection(
        self,
        identifier: str,
        other_attributes: _attributes_type | None,
    ) -> ProvRecord: ...
    def membership(self, collection: ProvRecord, entity: ProvEntity | str) -> ProvRecord: ...
    def plot(
        self,
        filename: str | None = ...,
        show_nary: bool = ...,
        use_labels: bool = ...,
        show_element_attributes: bool = ...,
        show_relation_attributes: bool = ...,
    ) -> None: ...
    wasGeneratedBy = generation
    used = usage
    wasStartedBy = start
    wasEndedBy = end
    wasInvalidatedBy = invalidation
    wasInformedBy = communication
    wasAttributedTo = attribution
    wasAssociatedWith = association
    actedOnBehalfOf = delegation
    wasInfluencedBy = influence
    wasDerivedFrom = derivation
    wasRevisionOf = revision
    wasQuotedFrom = quotation
    hadPrimarySource = primary_source
    alternateOf = alternate
    specializationOf = specialization
    mentionOf = mention
    hadMember = membership

class ProvDocument(ProvBundle):
    def __init__(
        self,
        records: Iterable[ProvRecord] | None = ...,
        namespaces: Iterable[Namespace] | None = ...,
    ) -> None: ...
    def is_document(self) -> bool: ...
    def is_bundle(self) -> bool: ...
    def has_bundles(self) -> bool: ...
    @property
    def bundles(self) -> Iterable[ProvBundle]: ...
    def flattened(self) -> ProvDocument: ...
    def unified(self) -> ProvDocument: ...
    def update(self, other: ProvDocument | ProvBundle) -> None: ...
    def add_bundle(self, bundle: ProvBundle, identifier: Incomplete | None = ...) -> None: ...
    def bundle(self, identifier: Identifier) -> ProvBundle: ...
    def serialize(
        self, destination: IO[Any] | None = ..., format: str = ..., **args: Any
    ) -> str | None: ...
    @staticmethod
    def deserialize(
        source: IO[Any] | str | None = ...,
        content: str | None = ...,
        format: str = ...,
        **args: Any,
    ) -> ProvDocument: ...

def sorted_attributes(element: ProvElement, attributes: List[str]) -> List[str]: ...



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0
"""TaskQueue."""

import queue
import threading
from typing import Callable, Optional

from .loghandler import _logger


class TaskQueue:
    """A TaskQueue class.

    Uses a first-in, first-out queue of tasks executed on a fixed number of
    threads.

    New tasks enter the queue and are started in the order received,
    as worker threads become available.

    If thread_count == 0 then tasks will be synchronously executed
    when add() is called (this makes the actual task queue behavior a
    no-op, but may be a useful configuration knob).

    The thread_count is also used as the maximum size of the queue.

    The threads are created during TaskQueue initialization.  Call
    join() when you're done with the TaskQueue and want the threads to
    stop.
    """

    in_flight: int = 0
    """The number of tasks in the queue."""

    def __init__(self, lock: threading.Lock, thread_count: int):
        """Create a new task queue using the specified lock and number of threads."""
        self.thread_count = thread_count
        self.task_queue: queue.Queue[Optional[Callable[[], None]]] = queue.Queue(
            maxsize=self.thread_count
        )
        self.task_queue_threads = []
        self.lock = lock
        self.error: Optional[BaseException] = None

        for _r in range(0, self.thread_count):
            t = threading.Thread(target=self._task_queue_func)
            self.task_queue_threads.append(t)
            t.start()

    def _task_queue_func(self) -> None:
        while True:
            task = self.task_queue.get()
            if task is None:
                return
            try:
                task()
            except BaseException as e:  # noqa: B036
                _logger.exception("Unhandled exception running task", exc_info=e)
                self.error = e
            finally:
                with self.lock:
                    self.in_flight -= 1

    def add(
        self,
        task: Callable[[], None],
        unlock: Optional[threading.Condition] = None,
        check_done: Optional[threading.Event] = None,
    ) -> None:
        """
        Add your task to the queue.

        The optional unlock will be released prior to attempting to add the
        task to the queue.

        If the optional "check_done" threading.Event's flag is set, then we
        will skip adding this task to the queue.

        If the TaskQueue was created with thread_count == 0 then your task will
        be synchronously executed.

        """
        if self.thread_count == 0:
            task()
            return

        with self.lock:
            self.in_flight += 1

        while True:
            try:
                if unlock is not None:
                    unlock.release()
                if check_done is not None and check_done.is_set():
                    with self.lock:
                        self.in_flight -= 1
                    return
                self.task_queue.put(task, block=True, timeout=3)
                return
            except queue.Full:
                pass
            finally:
                if unlock is not None:
                    unlock.acquire()

    def drain(self) -> None:
        """Drain the queue."""
        try:
            while not self.task_queue.empty():
                self.task_queue.get(True, 0.1)
        except queue.Empty:
            pass

    def join(self) -> None:
        """Wait for all threads to complete."""
        for _t in self.task_queue_threads:
            self.task_queue.put(None)
        for t in self.task_queue_threads:
            t.join()



#!/usr/bin/env python3
"""Setup for the reference implementation of the CWL standards."""
import os
import sys
import warnings

from setuptools import setup

if os.name == "nt":
    warnings.warn(
        "The CWL reference runner (cwltool) no longer supports running "
        "CWL workflows natively on MS Windows as its previous MS Windows "
        "support was incomplete and untested. Instead, please see "
        "https://pypi.org/project/cwltool/#ms-windows-users "
        "for instructions on running cwltool via "
        "Windows Subsystem for Linux 2 (WSL2). If don't need to execute "
        "CWL documents, then you can ignore this warning, but please "
        "consider migrating to https://pypi.org/project/cwl-utils/ "
        "for your CWL document processing needs.",
        stacklevel=1,
    )

SETUP_DIR = os.path.dirname(__file__)
README = os.path.join(SETUP_DIR, "README.rst")

NEEDS_PYTEST = {"pytest", "test", "ptr"}.intersection(sys.argv)
PYTEST_RUNNER = ["pytest-runner", "pytest-cov"] if NEEDS_PYTEST else []
USE_MYPYC = False
# To compile with mypyc, a mypyc checkout must be present on the PYTHONPATH
if len(sys.argv) > 1 and sys.argv[1] == "--use-mypyc":
    sys.argv.pop(1)
    USE_MYPYC = True
if os.getenv("CWLTOOL_USE_MYPYC", None) == "1":
    USE_MYPYC = True

if USE_MYPYC:
    mypyc_targets = [
        "cwltool/argparser.py",
        "cwltool/builder.py",
        "cwltool/checker.py",
        "cwltool/command_line_tool.py",
        # "cwltool/context.py",  # monkeypatching
        "cwltool/cwlrdf.py",
        "cwltool/docker_id.py",
        "cwltool/docker.py",
        "cwltool/udocker.py",
        "cwltool/errors.py",
        "cwltool/executors.py",
        "cwltool/factory.py",
        "cwltool/flatten.py",
        # "cwltool/__init__.py",
        "cwltool/job.py",
        "cwltool/load_tool.py",
        # "cwltool/loghandler.py",  # so we can monkeypatch the logger from tests
        # "cwltool/__main__.py",
        "cwltool/main.py",
        "cwltool/mutation.py",
        "cwltool/pack.py",
        "cwltool/pathmapper.py",
        "cwltool/process.py",
        "cwltool/procgenerator.py",
        # "cwltool/cwlprov/__init__.py",
        "cwltool/cwlprov/provenance_constants.py",
        "cwltool/cwlprov/provenance_profile.py",
        "cwltool/cwlprov/ro.py",
        # "cwltool/cwlprov/writablebagfile.py",  # WritableBag is having issues
        "cwltool/resolver.py",
        "cwltool/secrets.py",
        "cwltool/singularity.py",
        "cwltool/software_requirements.py",
        # "cwltool/stdfsaccess.py",  # StdFsAccess needs to be subclassable
        "cwltool/subgraph.py",
        "cwltool/update.py",
        "cwltool/utils.py",
        "cwltool/validate_js.py",
        "cwltool/workflow.py",
    ]

    from mypyc.build import mypycify  # type: ignore[import-untyped]

    opt_level = os.getenv("MYPYC_OPT_LEVEL", "3")
    ext_modules = mypycify(mypyc_targets, opt_level=opt_level)
else:
    ext_modules = []

setup(
    name="cwltool",
    description="Common workflow language reference implementation",
    long_description=open(README).read(),
    long_description_content_type="text/x-rst",
    author="Common workflow language working group",
    author_email="common-workflow-language@googlegroups.com",
    url="https://github.com/common-workflow-language/cwltool",
    download_url="https://github.com/common-workflow-language/cwltool",
    ext_modules=ext_modules,
    # platforms='',  # empty as is conveyed by the classifier below
    # license='',  # empty as is conveyed by the classifier below
    packages=[
        "cwltool",
        "cwltool.cwlprov",
        "cwltool.jshint",
        "cwltool.rdfqueries",
        "cwltool.schemas",
        "cwltool.tests",
        "cwltool.tests.checker_wf",
        "cwltool.tests.input_deps",
        "cwltool.tests.loop",
        "cwltool.tests.override",
        "cwltool.tests.reloc",
        "cwltool.tests.subgraph",
        "cwltool.tests.test_deps_env",
        "cwltool.tests.test_deps_env.modulefiles",
        "cwltool.tests.tmp1.tmp2.tmp3",
        "cwltool.tests.tmp4.alpha",
        "cwltool.tests.trs",
        "cwltool.tests.wf",
        "cwltool.tests.wf.generator",
        "cwltool.tests.wf.indir",
        "cwltool.tests.wf.operation",
    ],
    package_dir={"cwltool.tests": "tests"},
    include_package_data=True,
    install_requires=[
        "setuptools",
        "requests >= 2.6.1",  # >= 2.6.1 to workaround
        # https://github.com/ionrock/cachecontrol/issues/137
        "ruamel.yaml >= 0.16, < 0.19",
        "rdflib >= 4.2.2, < 7.1.0",
        "shellescape >= 3.4.1, < 3.9",
        "schema-salad >= 8.4.20230426093816, < 9",
        "prov == 1.5.1",
        "mypy-extensions",
        "psutil >= 5.6.6",
        "importlib_resources>=1.4;python_version<'3.9'",
        "coloredlogs",
        "pydot >= 1.4.1",
        "argcomplete",
        "pyparsing != 3.0.2",  # breaks --print-dot (pydot) https://github.com/pyparsing/pyparsing/issues/319
        "cwl-utils >= 0.32",
        "spython >= 0.3.0",
    ],
    extras_require={
        "deps": [
            "galaxy-tool-util>=22.1.2,!=23.0.1,!=23.0.2,!=23.0.3,!=23.0.4,!=23.0.5,<24.2",
            "galaxy-util <24.2",
        ],
    },
    python_requires=">=3.8, <4",
    use_scm_version=True,
    setup_requires=PYTEST_RUNNER + ["setuptools_scm>=8.0.4,<9"],
    test_suite="tests",
    tests_require=[
        "bagit >= 1.6.4, < 1.9",
        "pytest >= 6.2, < 8.3",
        "mock >= 2.0.0",
        "pytest-mock >= 1.10.0",
        "pytest-httpserver",
        "arcp >= 0.2.0",
    ],
    entry_points={"console_scripts": ["cwltool=cwltool.main:run"]},
    zip_safe=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Healthcare Industry",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: System :: Distributed Computing",
        "Topic :: Utilities",
    ],
)



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class FOAF(DefinedNamespace):
    account: URIRef
    accountName: URIRef
    accountServiceHomepage: URIRef
    age: URIRef
    based_near: URIRef
    birthday: URIRef
    currentProject: URIRef
    depiction: URIRef
    depicts: URIRef
    dnaChecksum: URIRef
    familyName: URIRef
    family_name: URIRef
    firstName: URIRef
    focus: URIRef
    fundedBy: URIRef
    geekcode: URIRef
    gender: URIRef
    givenName: URIRef
    givenname: URIRef
    holdsAccount: URIRef
    img: URIRef
    interest: URIRef
    knows: URIRef
    lastName: URIRef
    made: URIRef
    maker: URIRef
    member: URIRef
    membershipClass: URIRef
    myersBriggs: URIRef
    name: URIRef
    nick: URIRef
    page: URIRef
    pastProject: URIRef
    phone: URIRef
    plan: URIRef
    primaryTopic: URIRef
    publications: URIRef
    schoolHomepage: URIRef
    sha1: URIRef
    skypeID: URIRef
    status: URIRef
    surname: URIRef
    theme: URIRef
    thumbnail: URIRef
    tipjar: URIRef
    title: URIRef
    topic: URIRef
    topic_interest: URIRef
    workInfoHomepage: URIRef
    workplaceHomepage: URIRef
    Agent: URIRef
    Document: URIRef
    Group: URIRef
    Image: URIRef
    LabelProperty: URIRef
    OnlineAccount: URIRef
    OnlineChatAccount: URIRef
    OnlineEcommerceAccount: URIRef
    OnlineGamingAccount: URIRef
    Organization: URIRef
    Person: URIRef
    PersonalProfileDocument: URIRef
    Project: URIRef
    aimChatID: URIRef
    homepage: URIRef
    icqChatID: URIRef
    isPrimaryTopicOf: URIRef
    jabberID: URIRef
    logo: URIRef
    mbox: URIRef
    mbox_sha1sum: URIRef
    msnChatID: URIRef
    openid: URIRef
    weblog: URIRef
    yahooChatID: URIRef
